import logging
from typing import List, Dict, Tuple, Iterable, Union

import numpy as np
from datasets import load_dataset, DatasetDict, Dataset

from goodai.helpers.tokenizer_helper import get_sentence_punctuation_ids
from goodai.ltm.data.query_passage.example import QueryPassageExample
from transformers import PreTrainedTokenizer

from goodai.ltm.data.names import NameSource
from goodai.ltm.data.query_passage.data_source import BaseQueryPassageDataSource
from goodai.ltm.data.query_passage.qa_tok_entry import QATokenizedEntry


class QAQueryPassageDataSource(BaseQueryPassageDataSource):
    def __init__(self, random: np.random.RandomState, tokenizer: PreTrainedTokenizer,
                 qa_examples: List[dict],
                 max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int,
                 min_query_tokens=6, min_answer_start_leeway: int = 16, answer_leeway_min_tokens: int = 32):
        super().__init__()
        self.answer_leeway_min_tokens = answer_leeway_min_tokens
        self.min_answer_start_leeway = min_answer_start_leeway
        if max_query_tokens < min_query_tokens:
            raise ValueError(f'Expected max_query_tokens to be at least {min_query_tokens}')
        self.qa_examples = qa_examples
        self.min_query_tokens = min_query_tokens
        self.max_passage_tokens = max_passage_tokens
        self.min_passage_tokens = min_passage_tokens
        self.max_query_tokens = max_query_tokens
        self.tokenizer = tokenizer
        self.random = random
        self.tokenization_cache: Dict[int, QATokenizedEntry] = dict()
        self.punctuation_ids = get_sentence_punctuation_ids(tokenizer,
                                                            include_line_break=False, include_colon=True)

    @classmethod
    def _prepare_coqa_eval(cls, data: Dataset) -> Dataset:
        data = cls._add_id_column(data)

        def process_example(example):
            # Currently, use only the first question / answer and discard follow-up dialog
            example['question'] = example['questions'][0]
            example['answers'] = {'text': [example['answers']['input_text'][0]],
                                  'answer_start': [example['answers']['answer_start'][0]]}
            return example

        return data.map(process_example, remove_columns=['questions'])

    @classmethod
    def prepare_dataset(cls, ds_name: str):
        if ds_name == 'adversarial_qa':
            ds = load_dataset(ds_name, 'adversarialQA')
        else:
            ds = load_dataset(ds_name)
        if ds_name == 'coqa':
            # Use only coca_fiction
            coqa_fiction = ds.filter(lambda example: example['source'] in {'gutenberg', 'mctest'})
            train_test = coqa_fiction['train'].train_test_split(train_size=.9)
            ds = DatasetDict({
                'train': cls._prepare_coqa_eval(train_test['train']),
                'test': cls._prepare_coqa_eval(train_test['test']),
            })
        train_data = ds['train']
        test_data = ds.get('test')
        if test_data is None:
            test_data = ds['validation']
        train_data = cls._filter_valid(cls._process(train_data, ds_name))
        test_data = cls._filter_valid(cls._process(test_data, ds_name))
        return {
            'train': list(train_data),
            'test': list(test_data),
        }

    @classmethod
    def _add_id_column(cls, data):
        return data.add_column('id', [str(i) for i in range(data.num_rows)])

    @classmethod
    def _filter_valid(cls, entries: List[dict]):
        def _is_valid(entry: dict) -> bool:
            answers = entry['answers']
            a_text = answers['text']
            return len(a_text) > 0

        return [e for e in entries if _is_valid(e)]

    @classmethod
    def _process(cls, entries: Iterable[dict], ds_name: str) -> List[dict]:
        def _extend(e: dict):
            new_dict = dict(e)
            new_dict['ds_name'] = ds_name
            return new_dict

        return [_extend(entry) for entry in entries]

    @classmethod
    def create_data_sources(cls, random: np.random.RandomState, ds_name: str, tokenizer: PreTrainedTokenizer,
                            max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int) ->\
            Tuple['QAQueryPassageDataSource', 'QAQueryPassageDataSource']:
        data = cls.prepare_dataset(ds_name)
        train_examples = data['train']
        test_examples = data['test']
        train_ds = cls(random, tokenizer, train_examples, max_query_tokens, min_passage_tokens, max_passage_tokens)
        test_ds = cls(random, tokenizer, test_examples, max_query_tokens, min_passage_tokens, max_passage_tokens)
        return train_ds, test_ds,

    def get_tokenization(self, index: int) -> QATokenizedEntry:
        tok_entry = self.tokenization_cache.get(index)
        if tok_entry is None:
            index = self.random.randint(0, len(self.qa_examples))
            qa_example: dict = self.qa_examples[index]
            tok_entry = QATokenizedEntry.from_example(qa_example, self.tokenizer)
            self.tokenization_cache[index] = tok_entry
        return tok_entry

    def get_query_token_ids(self, tok_entry: QATokenizedEntry, ends_with_name_p=0.75, use_names_p=0.95,
                            query_noise_p=0.75):
        r = self.random
        name2 = None
        if r.uniform() < use_names_p:
            name1, name2 = tuple(NameSource.get_instance().sample_first_names(r, 2))
        else:
            name1 = r.choice(['User', 'Player'])
        has_query_noise = self.random.uniform() < query_noise_p
        name_lead = ('\n' if has_query_noise else '') + name1 + ':'
        name1_ids = self.tokenizer.encode(name_lead, add_special_tokens=False)
        token_ids = name1_ids + tok_entry.question_token_ids
        ends_with_name = r.uniform() < ends_with_name_p
        if ends_with_name:
            if name2 is None:
                name2 = r.choice(['AI', 'Assistant', 'Agent', 'NPC'])
            name2_ids = self.tokenizer.encode('\n' + name2 + ':', add_special_tokens=False)
            token_ids += name2_ids
        if has_query_noise:
            min_token_ids = min(self.max_query_tokens, max(self.min_query_tokens, len(token_ids)))
            num_tokens = r.randint(min_token_ids, self.max_query_tokens + 1)
            full_sequence = tok_entry.content_token_ids + token_ids
            num_noise_tokens = max(0, num_tokens - len(token_ids))
            return full_sequence[-num_tokens:], num_noise_tokens,
        else:
            return token_ids[-self.max_query_tokens:], 0,

    def get_adjusted_answer_start(self, context_ids: List[int], answer_start: int, num_chunk_tokens: int):
        answer_start_leeway = self.min_answer_start_leeway + max(0, num_chunk_tokens - self.answer_leeway_min_tokens)
        found_index = max(0, answer_start - answer_start_leeway)
        for i in range(1, answer_start_leeway + 1):
            token_index = answer_start - i
            if token_index < 0:
                found_index = 0
                break
            token_id = context_ids[token_index]
            if token_id in self.punctuation_ids:
                found_index = token_index + 1
                break
        adjustment = answer_start - found_index
        extra_leeway = max(0, num_chunk_tokens - adjustment - self.answer_leeway_min_tokens)
        if extra_leeway > 0:
            offset = self.random.randint(0, extra_leeway + 1)
            found_index = max(0, found_index - offset)
        return found_index

    def get_passage_token_ids(self, tok_entry: QATokenizedEntry, is_match: bool, non_answer: bool,
                              non_answer_exclude_last_n: int):
        r = self.random
        context_ids = tok_entry.content_token_ids
        answer_start = tok_entry.answer_seq_index
        num_tokens = r.randint(self.min_passage_tokens, self.max_passage_tokens + 1)
        adj_answer_start = self.get_adjusted_answer_start(context_ids, answer_start, num_tokens)
        if is_match:
            passage_start = adj_answer_start
        elif non_answer:
            num_tokens_right = len(context_ids) - (answer_start + self.max_passage_tokens) - non_answer_exclude_last_n
            num_tokens_left = answer_start - self.max_passage_tokens
            if num_tokens_right >= num_tokens_left:
                p_from = answer_start + self.max_passage_tokens
                p_to = len(context_ids) - non_answer_exclude_last_n - num_tokens + 1
            else:
                p_from = 0
                p_to = answer_start - self.max_passage_tokens - num_tokens + 1
            if p_from >= p_to:
                return None
            passage_start = r.randint(p_from, p_to)
        else:
            p_to = len(context_ids) - num_tokens + 1
            if p_to < 1:
                return None
            passage_start = r.randint(0, p_to)
        passage_end = passage_start + num_tokens
        result = context_ids[passage_start:passage_end]
        if len(result) < self.min_passage_tokens:
            return None
        return result

    def sample_item(self, pos_index: int, is_match: bool, use_different_p=0.5) -> QueryPassageExample:
        r = self.random
        n = len(self.qa_examples)
        for attempt in range(100):
            pos_tok_entry = self.get_tokenization(pos_index)
            if pos_tok_entry.answer_seq_index == -1:
                logging.warning(f'Did not find location of answer excerpt in example with id {pos_tok_entry.e_id}')
                pos_index = r.randint(0, n)
                continue
            query_token_ids, nqn = self.get_query_token_ids(pos_tok_entry)
            if is_match:
                passage_tok_entry = pos_tok_entry
                non_answer = False
            else:
                use_different = r.uniform() < use_different_p
                if use_different:
                    neg_index = r.randint(0, n)
                    neg_tok_entry = self.get_tokenization(neg_index)
                    non_answer = neg_index == pos_index
                    passage_tok_entry = neg_tok_entry
                else:
                    non_answer = True
                    passage_tok_entry = pos_tok_entry
            passage_token_ids = self.get_passage_token_ids(passage_tok_entry, is_match, non_answer, nqn)
            if passage_token_ids is None:
                if is_match or non_answer:
                    pos_index = r.randint(0, n)
                continue
            return QueryPassageExample(query_token_ids, passage_token_ids, is_match)

        raise SystemError('Unable to find suitable qa_example!')

    def sample_item_indexes(self, total_items: int, sample_count: int) -> Union[List[int], np.ndarray]:
        replace = sample_count > total_items
        return self.random.choice(range(total_items), size=sample_count, replace=replace)

    def sample_items(self, count: int, approx_positive_fraction: float = 0.5) -> List[QueryPassageExample]:
        rnd_samples = self.random.uniform(size=count)
        pos_indexes = self.sample_item_indexes(len(self.qa_examples), count)
        is_match = rnd_samples <= approx_positive_fraction
        return [self.sample_item(pos_indexes[i], is_match[i]) for i, _ in enumerate(range(count))]



import itertools
import math

import torch
from torch import nn

from goodai.ltm.embeddings.contrast_classifier import ContrastClassifier
from goodai.ltm.embeddings.trainable import TrainableEmbeddingModel
from goodai.ltm.reranking.prob_model import BaseQueryPassageProbModel


class EmbeddingQueryPassageProbModel(nn.Module, BaseQueryPassageProbModel):
    def __init__(self, emb_model: TrainableEmbeddingModel):
        super(EmbeddingQueryPassageProbModel, self).__init__()
        self.emb_model = emb_model
        scale = math.sqrt(emb_model.get_embedding_dim())
        self.classifier = ContrastClassifier(scale)

    def forward(self, query_input_ids: torch.Tensor, query_attention_mask: torch.Tensor,
                query_token_lengths: torch.Tensor,
                passage_input_ids: torch.Tensor, passage_attention_mask: torch.Tensor) -> torch.Tensor:
        rk = self.emb_model.get_retrieval_emb(query_input_ids, token_lengths=query_token_lengths,
                                              attention_mask=query_attention_mask)
        sk = self.emb_model.get_storage_emb(passage_input_ids, passage_attention_mask)
        return self.classifier(rk, sk)

    def get_lm_parameters(self):
        return self.emb_model.get_lm_parameters()

    def get_added_parameters(self):
        km_added_parameters = self.emb_model.get_added_parameters()
        c_parameters = self.classifier.parameters()
        return itertools.chain(km_added_parameters, c_parameters)




import torch
from typing import List, Tuple
from tqdm import tqdm
from goodai.ltm.embeddings.base import BaseTextEmbeddingModel
from goodai.ltm.reranking.base import BaseTextMatchingModel
from goodai.modules.loss import EmbCrossProbLossModel


class EmbeddingBasedMatchingModel(BaseTextMatchingModel):
    def __init__(self, emb_model: BaseTextEmbeddingModel, dist_param: float):
        super().__init__()
        self.dist_param = dist_param
        self.emb_model = emb_model

    def predict(self, sentences: List[Tuple[str, str]], batch_size: int = 32,
                show_progress_bar: bool = False, add_special_tokens: bool = True) -> List[float]:
        with torch.no_grad():
            rng = range(0, len(sentences), batch_size)
            if show_progress_bar:
                rng = tqdm(rng, desc='Encoding QPs', unit='batch')
            result: List[float] = []
            for b0 in rng:
                b_sentences = sentences[b0:b0 + batch_size]
                queries, passages = zip(*b_sentences)
                rk = self.emb_model.encode_queries(queries, batch_size=batch_size, show_progress_bar=False,
                                                   convert_to_tensor=True)
                rk = rk.detach()
                sk = self.emb_model.encode_corpus(passages, batch_size=batch_size, show_progress_bar=False,
                                                  convert_to_tensor=True)
                sk = sk.detach()
                # rk: (batch_size, num_r_emb, emb_dim,)
                # sk: (batch_size, num_s_emb, emb_dim,)
                distances = torch.cdist(rk, sk)
                # distances: (batch_size, num_r_emb, num_s_emb,)
                batch_size = distances.size(0)
                distances = distances.view(batch_size, -1)
                min_distances = torch.amin(distances, dim=1, keepdim=True)
                prob_t = EmbCrossProbLossModel.get_prob(min_distances, self.dist_param)
                partial_result = prob_t.squeeze(1).tolist()
                result.extend(partial_result)
            return result

    def get_info(self):
        return f'{self.__class__}: EmbModel {self.emb_model.get_info()}'



from enum import Enum
from typing import Optional, List, Any

from transformers import PreTrainedTokenizer


class ChunkExpansionLimitType(Enum):
    SENTENCE = 0,
    LINE = 1,
    PARAGRAPH = 2,
    SECTION = 3

    @staticmethod
    def batch_encode(tokenizer: PreTrainedTokenizer, texts: List[str]):
        tok = tokenizer.batch_encode_plus(texts, add_special_tokens=False, return_attention_mask=False)
        return tok['input_ids']

    @staticmethod
    def distinct(a_list: List[List[Any]]) -> List[List[Any]]:
        tuples = [tuple(x) for x in a_list]
        tuples = list(set(tuples))
        return [list(x) for x in tuples]

    def get_token_ids(self, tokenizer: PreTrainedTokenizer) -> List[List[int]]:
        if self == self.SENTENCE:
            return self.batch_encode(tokenizer, ['.', '?', '!', '???', '!!!'])
        elif self == self.LINE:
            cr_id = tokenizer.encode('\r', add_special_tokens=False)
            nl_id = tokenizer.encode('\n', add_special_tokens=False)
            result = self.batch_encode(tokenizer, ['\n', '\r\n']) + \
                [cr_id + nl_id]
            return self.distinct(result)
        elif self == self.PARAGRAPH:
            cr_id = tokenizer.encode('\r', add_special_tokens=False)
            nl_id = tokenizer.encode('\n', add_special_tokens=False)
            result = self.batch_encode(tokenizer, ['\n\n', '\r\n\r\n']) + \
                [nl_id + nl_id] + [cr_id + nl_id + cr_id + nl_id]
            return self.distinct(result)
        elif self == self.SECTION:
            return []
        else:
            raise ValueError(f'Unhandled limit type: {self}')


class ChunkExpansionConfig:
    def __init__(self, min_extra_side_tokens: int = 0, max_extra_side_tokens: int = 24,
                 limit_type: ChunkExpansionLimitType = ChunkExpansionLimitType.SENTENCE):
        self.limit_type = limit_type
        self.min_extra_side_tokens = min_extra_side_tokens
        self.max_extra_side_tokens = max_extra_side_tokens

    @classmethod
    def for_sentence(cls, max_extra_side_tokens: int = 24):
        return cls(max_extra_side_tokens=max_extra_side_tokens, limit_type=ChunkExpansionLimitType.SENTENCE)

    @classmethod
    def for_line_break(cls, min_extra_side_tokens: int = 16, max_extra_side_tokens: int = 64):
        return cls(min_extra_side_tokens=min_extra_side_tokens,
                   max_extra_side_tokens=max_extra_side_tokens,
                   limit_type=ChunkExpansionLimitType.LINE)

    @classmethod
    def for_paragraph(cls, max_extra_side_tokens: int = 192):
        return cls(max_extra_side_tokens=max_extra_side_tokens, limit_type=ChunkExpansionLimitType.PARAGRAPH)

    @classmethod
    def for_section(cls, max_extra_side_tokens: int = 1024):
        return cls(max_extra_side_tokens=max_extra_side_tokens, limit_type=ChunkExpansionLimitType.SECTION)

    @classmethod
    def for_chunk(cls):
        return cls(max_extra_side_tokens=0, limit_type=ChunkExpansionLimitType.SECTION)


class TextMemoryConfig:
    chunk_capacity: int
    """
    The capacity of a chunk, in tokens.
    """

    queue_capacity: int
    """
    The capacity of the memory's chunk queue, i.e. the maximum number of chunks it can hold.
    """

    reranking_k_factor: float
    """
    When a reranking mechanism is available, the value of k passed to the retrieval function is
    multiplied by reranking_k_factor and the resulting number of chunks is retrieved from
    the chunk store before applying the reranking procedure.
    """

    max_query_length: Optional[int]
    """
    The maximum length of queries used to produce embeddings. The left side of queries
    exceeding this maximum length are truncated (after rewriting). 
    """

    chunk_overlap_fraction: float
    """
    The fraction of a chunk's length that overlaps with the previous and next chunk.
    This can be a value from 0 to 0.5.
    """

    redundancy_overlap_threshold: float
    """
    The fraction of a retrieved passage's length that causes it to be considered 
    redundant if it overlaps with other passages that are a better match to a query.
    """

    chunk_expansion_config: ChunkExpansionConfig
    """
    The chunk expansion configuration.
    """

    def __init__(self):
        self.max_query_length = 40  # Tokens
        self.chunk_capacity = 24  # Tokens
        self.queue_capacity = 64000  # Chunks
        self.reranking_k_factor = 10.0
        self.chunk_overlap_fraction = 0.5  # 0 to 0.5
        self.redundancy_overlap_threshold = 0.75
        self.chunk_expansion_config = ChunkExpansionConfig(max_extra_side_tokens=self.chunk_capacity)



import torch
import torch.nn.utils.rnn as R
from typing import Set, List, Tuple, Optional
from transformers import PreTrainedTokenizer


def get_pad_token_id(tokenizer: PreTrainedTokenizer):
    p_tid = tokenizer.pad_token_id
    if p_tid is None:
        tokens = tokenizer.tokenize(' ')
        if len(tokens) == 0:
            tokens = tokenizer.tokenize('-')
            if len(tokens) == 0:
                raise Exception('Unable to find suitable padding token')
        token_ids = tokenizer.convert_tokens_to_ids(tokens)
        p_tid = token_ids[0]
    return p_tid


def get_eos_token_id(tokenizer: PreTrainedTokenizer):
    eos_id = tokenizer.eos_token_id
    if eos_id is None:
        eos_id = get_pad_token_id(tokenizer)
    return eos_id


def get_sentence_punctuation_ids(tokenizer: PreTrainedTokenizer, include_line_break=False,
                                 include_colon=False) -> Set[int]:
    def _update(text: str, r_set: set):
        token_ids = tokenizer.encode(text, add_special_tokens=False)
        if len(token_ids) == 1:
            r_set.add(token_ids[0])

    result = set()
    _update('.', result)
    _update('!', result)
    _update('!!', result)
    _update('!!!', result)
    _update('?', result)
    _update('??', result)
    _update('???', result)
    if include_line_break:
        _update('\n', result)
    if include_colon:
        _update(':', result)
    return result


def get_token_index(token_ids: List[int], token_offsets: List[Tuple[int, int]], offset: int):
    for i, (token_id, offset_pair) in enumerate(zip(token_ids, token_offsets)):
        o_from, o_to = offset_pair
        if o_from <= offset <= o_to:
            return i
    return -1


def get_model_inputs(input_ids_list: List[List[int]], pad_id: int, device: torch.device,
                     min_seq_len: Optional[int] = None, prefix: str = '',
                     tokenizer: PreTrainedTokenizer = None, return_token_lengths: bool = False):
    # Note: padding on the right
    max_seq_len = max(len(seq) for seq in input_ids_list)
    if min_seq_len is None:
        min_seq_len = max_seq_len
    else:
        min_seq_len = max(max_seq_len, min_seq_len)
    input_ids_list = [ids + [pad_id] * (min_seq_len - len(ids)) for ids in input_ids_list]
    input_ids = torch.as_tensor(input_ids_list, dtype=torch.int64, device=device)
    attention_mask = (input_ids != pad_id).float()
    result = {
        prefix + 'input_ids': input_ids,
        prefix + 'attention_mask': attention_mask,
    }
    if return_token_lengths:
        def _lengths(seq: List[int]):
            tokens = tokenizer.convert_ids_to_tokens(seq)
            return [len(t) for t in tokens]

        if tokenizer is None:
            raise ValueError('tokenizer cannot be None when token lengths are requested')
        token_lengths = [_lengths(seq) for seq in input_ids_list]
        result[prefix + 'token_lengths'] = torch.as_tensor(token_lengths, dtype=torch.long, device=device)
    return result


def get_attention_after_token(input_ids: torch.Tensor, attention_mask: torch.Tensor, token_lengths: torch.Tensor,
                              token_id: int, exclude_last_n_chars: int, device: torch.device) -> torch.Tensor:
    att_lengths = token_lengths * attention_mask
    inc_lengths = torch.flip(torch.cumsum(torch.flip(att_lengths, dims=(1,)), dim=1), dims=(1,))
    considered: torch.Tensor = torch.greater(inc_lengths, exclude_last_n_chars)
    matches_token: torch.Tensor = torch.eq(input_ids, token_id) & considered
    index_range = torch.arange(0, input_ids.size(1), device=device).unsqueeze(0)
    indices = torch.where(matches_token, index_range, -1)
    max_indices, _ = torch.max(indices, dim=1, keepdim=True)
    at_att_mask = torch.greater(index_range, max_indices).float()
    return at_att_mask




from dataclasses import dataclass
from typing import List, Dict, Tuple, Iterable, Union

import numpy as np
from datasets import load_dataset

from goodai.helpers.tokenizer_helper import get_sentence_punctuation_ids
from goodai.ltm.data.query_passage.example import QueryPassageExample
from transformers import PreTrainedTokenizer

from goodai.ltm.data.names import NameSource
from goodai.ltm.data.query_passage.data_source import BaseQueryPassageDataSource


@dataclass
class SharcEntry:
    questionIds: List[int]
    snippetIds: List[int]
    scenarioIds: List[int]

    @staticmethod
    def from_example(example: dict, tokenizer: PreTrainedTokenizer) -> 'SharcEntry':
        question = example['question']
        scenario = example['scenario']
        snippet = example['snippet']
        snippet = snippet.strip()
        while len(snippet) > 0 and snippet[0] == '#':
            snippet = snippet[1:]
        question_ids = tokenizer.encode(' ' + question, add_special_tokens=False)
        scenario_ids = tokenizer.encode(' ' + scenario, add_special_tokens=False)
        snippet_ids = tokenizer.encode(snippet, add_special_tokens=False)
        return SharcEntry(question_ids, snippet_ids, scenario_ids)


class SharcQueryPassageDataSource(BaseQueryPassageDataSource):
    def __init__(self, random: np.random.RandomState, tokenizer: PreTrainedTokenizer,
                 qa_examples: List[dict],
                 max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int,
                 min_query_tokens=6):
        super().__init__()
        if max_query_tokens < min_query_tokens:
            raise ValueError(f'Expected max_query_tokens to be at least {min_query_tokens}')
        self.qa_examples = qa_examples
        self.min_query_tokens = min_query_tokens
        self.max_passage_tokens = max_passage_tokens
        self.min_passage_tokens = min_passage_tokens
        self.max_query_tokens = max_query_tokens
        self.tokenizer = tokenizer
        self.random = random
        self.tokenization_cache: Dict[int, SharcEntry] = dict()
        self.punctuation_ids = get_sentence_punctuation_ids(tokenizer,
                                                            include_line_break=False, include_colon=True)

    @classmethod
    def prepare_dataset(cls, ds_name: str):
        ds = load_dataset(ds_name)
        train_data = ds['train']
        test_data = ds.get('test')
        if test_data is None:
            test_data = ds['validation']
        train_data = cls._filter_valid(cls._process(train_data, ds_name))
        test_data = cls._filter_valid(cls._process(test_data, ds_name))
        return {
            'train': list(train_data),
            'test': list(test_data),
        }

    @classmethod
    def _add_id_column(cls, data):
        return data.add_column('id', [str(i) for i in range(data.num_rows)])

    @classmethod
    def _filter_valid(cls, entries: List[dict]):
        def _is_valid(entry: dict) -> bool:
            snippet = entry.get('snippet')
            return snippet is not None and len(snippet) > 0

        return [e for e in entries if _is_valid(e)]

    @classmethod
    def _process(cls, entries: Iterable[dict], ds_name: str) -> List[dict]:
        def _extend(e: dict):
            new_dict = dict(e)
            new_dict['ds_name'] = ds_name
            return new_dict

        return [_extend(entry) for entry in entries]

    @classmethod
    def create_data_sources(cls, random: np.random.RandomState, ds_name: str, tokenizer: PreTrainedTokenizer,
                            max_query_tokens: int, min_passage_tokens: int, max_passage_tokens: int) ->\
            Tuple['SharcQueryPassageDataSource', 'SharcQueryPassageDataSource']:
        data = cls.prepare_dataset(ds_name)
        train_examples = data['train']
        test_examples = data['test']
        train_ds = cls(random, tokenizer, train_examples, max_query_tokens, min_passage_tokens, max_passage_tokens)
        test_ds = cls(random, tokenizer, test_examples, max_query_tokens, min_passage_tokens, max_passage_tokens)
        return train_ds, test_ds,

    def get_tokenization(self, index: int) -> SharcEntry:
        tok_entry = self.tokenization_cache.get(index)
        if tok_entry is None:
            index = self.random.randint(0, len(self.qa_examples))
            qa_example: dict = self.qa_examples[index]
            tok_entry = SharcEntry.from_example(qa_example, self.tokenizer)
            self.tokenization_cache[index] = tok_entry
        return tok_entry

    def get_query_token_ids(self, tok_entry: SharcEntry, ends_with_name_p=0.75, use_names_p=0.95,
                            query_noise_p=0.75):
        r = self.random
        name2 = None
        if r.uniform() < use_names_p:
            name1, name2 = tuple(NameSource.get_instance().sample_first_names(r, 2))
        else:
            name1 = r.choice(['User', 'Player'])
        has_query_noise = self.random.uniform() < query_noise_p
        name_lead = ('\n' if has_query_noise else '') + name1 + ':'
        name1_ids = self.tokenizer.encode(name_lead, add_special_tokens=False)
        token_ids = name1_ids + tok_entry.questionIds
        ends_with_name = r.uniform() < ends_with_name_p
        if ends_with_name:
            if name2 is None:
                name2 = r.choice(['AI', 'Assistant', 'Agent', 'NPC'])
            name2_ids = self.tokenizer.encode('\n' + name2 + ':', add_special_tokens=False)
            token_ids += name2_ids
        if has_query_noise and len(tok_entry.scenarioIds) > 1:
            min_token_ids = min(self.max_query_tokens, max(self.min_query_tokens, len(token_ids)))
            num_tokens = r.randint(min_token_ids, self.max_query_tokens + 1)
            name_scenario_ids = name1_ids + tok_entry.scenarioIds
            full_sequence = name_scenario_ids + token_ids
            num_noise_tokens = max(0, num_tokens - len(token_ids))
            return full_sequence[-num_tokens:], num_noise_tokens,
        else:
            return token_ids[-self.max_query_tokens:], 0,

    def get_passage_token_ids(self, tok_entry: SharcEntry):
        r = self.random
        context_ids = tok_entry.snippetIds
        excess = len(context_ids) - self.max_passage_tokens
        if excess > 0:
            idx = r.randint(0, excess + 1)
            return context_ids[idx:idx + self.max_passage_tokens]
        else:
            return context_ids

    def sample_item(self, pos_index: int, is_match: bool) -> QueryPassageExample:
        r = self.random
        pos_tok_entry = self.get_tokenization(pos_index)
        query_token_ids, nqn = self.get_query_token_ids(pos_tok_entry)
        if is_match:
            passage_tok_entry = pos_tok_entry
        else:
            neg_index = None
            n = len(self.qa_examples)
            for attempt in range(100):
                neg_index = r.randint(0, n)
                if neg_index == pos_index:
                    continue
                break
            if neg_index is None:
                raise SystemError('Unable to find suitable example!')
            neg_tok_entry = self.get_tokenization(neg_index)
            passage_tok_entry = neg_tok_entry
        passage_token_ids = self.get_passage_token_ids(passage_tok_entry)
        return QueryPassageExample(query_token_ids, passage_token_ids, is_match)

    def sample_item_indexes(self, total_items: int, sample_count: int) -> Union[List[int], np.ndarray]:
        replace = sample_count > total_items
        return self.random.choice(range(total_items), size=sample_count, replace=replace)

    def sample_items(self, count: int, approx_positive_fraction: float = 0.5) -> List[QueryPassageExample]:
        rnd_samples = self.random.uniform(size=count)
        pos_indexes = self.sample_item_indexes(len(self.qa_examples), count)
        is_match = rnd_samples <= approx_positive_fraction
        return [self.sample_item(pos_indexes[i], is_match[i]) for i, _ in enumerate(range(count))]



import codecs
import os
import unittest
from pathlib import Path
from goodai.ltm.data.query_passage.auto_data_source import AutoQueryPassageDataSource
from transformers import PreTrainedTokenizer, AutoTokenizer
import numpy as np
from goodai.helpers.html_helper import text_to_html


class TestQueryPassageDataSource(unittest.TestCase):
    def test_show_examples(self):
        tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained('distilroberta-base')
        random = np.random.RandomState(1)
        ds_names = ['wikianswers', 'sharc', 'adversarial_qa', 'coqa', 'squad_v2', 'wiki']
        # ds_names = ['coqa', 'squad_v2']
        for ds_name in ds_names:
            print(f'Processing {ds_name}...')
            self._show_examples(random, ds_name, tokenizer)

    @staticmethod
    def _show_examples(random: np.random.RandomState, ds_name: str, tokenizer: PreTrainedTokenizer):
        train_ds, test_ds = AutoQueryPassageDataSource.create(ds_name, random, tokenizer,
                                                              max_query_tokens=40, min_passage_tokens=24,
                                                              max_passage_tokens=36)

        # Sample 100 examples from the training data source
        examples = train_ds.sample_items(100, approx_positive_fraction=0.1)

        # Sort examples by match value
        examples.sort(key=lambda x: x.match, reverse=True)

        # Generate HTML table
        table_rows = ['<tr><th>Match</th><th>Query</th><th>Passage</th></tr>']
        for example in examples:
            match_str = 'Yes' if example.match else 'No'
            query_str = tokenizer.decode(example.queryIds, skip_special_tokens=True)
            passage_str = tokenizer.decode(example.passageIds, skip_special_tokens=True)
            query_html = text_to_html(query_str)
            passage_html = text_to_html(passage_str)
            table_rows.append(f'<tr><td>{match_str}</td><td>{query_html}</td>' +
                              f'<td>{passage_html}</td></tr>')
        html_table = '<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />\n' +\
                     '<html><body>' +\
                     '<table border=1>' + ''.join(table_rows) + '</table>' +\
                     '</body></html>'

        # Save HTML file to local data directory
        data_dir = Path('data')
        os.makedirs(data_dir, exist_ok=True)
        html_path = str(data_dir / f'qp_datasource_{ds_name}_examples.html')
        with codecs.open(html_path, 'w', 'utf-8') as f:
            f.write(html_table)



import json
import os
import string
from abc import abstractmethod, ABC
from textwrap import dedent
from typing import List

import openai


DEFAULT_QUERY_PROMPT_TEMPLATE = dedent("""\
    You will be provided a json document containing an "unprocessed" entry. 
    You will rewrite only the unprocessed entry into a query for a retrieval system. 
    Rewrite the unprocessed entry according to the instructions and examples below and store the rewritten query as a string called "processed". 
    - The entry contains a dialog with replies prefixed by the speaker's name. Keep track of the person making a request and the person receiving a request. Be careful to attribute requests to the right person. 
    { "unprocessed": "Joe: Do you know the password? Ben: No, I don't. But what I do need help with right now, if you're willing to lend a hand, has to do with Lucas Fern." } 
    { "processed": "Can Joe help Ben with Lucas Fern?" }
    - The entry contains replies prefixed by the speaker's name. Focus on the last entry.
    { "unprocessed": "Alice: So, I was thinking about how we can make this place more beautiful. Maybe you have some ideas? Gary: What is the password?" }
    { "processed": "What is the password?"}
    - Remove ellipsis, including pronouns and tacit information.
    { "unprocessed": "Jake Anderson: Have you seen the new F-150? James: Do you like it?"}
    { "processed": "Does Jake like the Ford F-150?" }
    - Adapt information based on qualifiers. Discard standalone qualifier sentences.
    { "unprocessed": "Ben: Tell me more. James: I am not sure about what I am to tell you next... Cows in Austria are violet."}
    { "processed": "James is not sure about it, but claims that cows in Austria are violet." }
    - Make subjective information objective. E.g.,
    { "unprocessed": "Emma: I like painting. What do you like? James: I think the best sport is cycling. I particularly enjoy Eddy Merckx."}
    { "processed": "James' favorite sport is cycling and James' favorite cyclist is Eddy Merckx." }
    - Wherever possible, include the answers to "who", "what", "when" and "where" in the processed . 
    { "unprocessed": "Nicholas: The Festival of the Stars is a yearly event that celebrates the celestial bodies and their influence on Aradia. Chris: It is a time of joy for all who attend!" }
    { "processed" : "Chris thinks the Festival of the Stars in Aradia is a time of joy for all who attend." }
    - Summarize wordy narratives into concise statements of relevant information.
    { "unprocessed": "Annie: What's the matter? John: I don't know where to begin... The guards are harassing me. It's awful... I don't know what to do. Just the other day it was very bad." }
    { "processed": "The guards are harassing John." }
    - When possible, summarize wordy entries into questions for relevant information.
    { "unprocessed": "Simon: Did you like the markets? Cecile: Yes, the seafood is good in Sicily and fresh vegetables are tasty and cheap. Why don't we live there... I could spend hours in those food markets... Do you remember the name of the old market in Palermo?"}
    { "processed": "What is the name of the old food market in Palermo?" }
    - Produce a complete sentence or query. Focus on the last reply but use information from preceding dialog if necessary.
    { "unprocessed": "Petr: Drones are cool. Jarda: Totally." }
    { "processed": "Jarda agrees that drones are cool." }
    - Don't ask questions that are answered in the dialog. Do ask questions for new information.
    { "unprocessed": "Ben: There is some food in the fridge. Jerry: I see. Will Johnny come over?" } 
    { "processed": "Will Johnny come over?" }

    Now rewrite the unprocessed entry. Keep track of the speakers in the dialog, but only mention them when necessary. 
    Don't ask questions that are answered in the dialog. Do ask for new information.
    { "unprocessed": "$query"}""")

DEFAULT_MEMORY_PROMPT_TEMPLATE = dedent("""\
    You will be provided a json document containing a "context" entry and an "unprocessed" entry. 
    You will rewrite only the unprocessed entry into individual facts that can be indexed easily by a retrieval system. 
    Rewrite the unprocessed entry according to the instructions below and store the rewritten facts as entries in a json array called "processed". 
    Make sure to split sentences into multiple rewritten facts if there is too much information in the "unprocessed" sentence.
    If you encounter a question, do not include it in the rewritten facts. Do not repeat facts or sentences mentioned in "context". Moreover:
    - remove ellipsis, including pronouns and tacit information.Example input/ output:
    { "context": "James: Have you seen the new F-150?\nJane: Sure, that thing is a beast, I wish I had it.\n",  "unprocessed": "James: I like it a lot."}
    { "processed": { "1": "James likes Ford F-150 a lot."} }
    - adapt information based on qualifiers. Discard standalone qualifier sentences.
    { "context": "James: I am not sure about what I am to tell you next.\nJane: Yes?\n", "unprocessed": "James: Cows in Austria are violet."}
    { "processed": { "1": "James is not sure about it, but claims that cows in Austria are violet."} }
    - split sentences with multiple unrelated pieces of information into separate, atomic facts. E.g.
    { "context": "Marry: What are your favorite things, James?", "unprocessed": "James: My favorite color is yellow and my favorite animal is elephant."}
    { "processed": { "1": "James' favorite color is yellow.", "2": "James' favorite animal is elephant."} }
    - make subjective information objective.E.g.
    { "context": "Jane: I love painting. It's the best thing.\n James: I like painting quite a bit too. But there is something else I like.\n", "unprocessed": "James: the best sport is cycling. I particularly enjoy Eddy Merckx."}
    { "processed": { "1": "James' favorite sport is cycling.", "2": "James' favorite cyclist is Eddy Merckx."} }
    - wherever possible, include the answers to "who", "what", "when" and "where" in the processed facts. 
    { "context": "There is a powerful and ancient magic that can bring people back from the dead, but it is forbidden and considered taboo by most inhabitants of Aradia.\nThe Festival of the Stars is a yearly event that celebrates the celestial bodies and their influence on Aradia.\n", "unprocessed": "It is a time of celebration and joy for all who attend." }
    { "processed" : { "1": "The Festival of the Stars in Aradia is a time of celebration and joy for all who attend."} }
    { "context": "$context" }
    { "unprocessed": "$passage" }""")

DEFAULT_QUERY_SYSTEM_MESSAGE = DEFAULT_MEMORY_SYSTEM_MESSAGE \
    = """You are a helpful assistant that rewrites queries and text passages for a text retrieval system."""


class BaseRewriteModel(ABC):
    """
    Abstract base class for query and memory rewrite models.

    Query and memory rewriting is used to disambiguate queries and passages and make them more self-contained.

    For example, the statement "It's the best!", referring to strawberry ice cream and uttered by John,
    can be rewritten as "John thinks that strawberry ice cream is the best."
    """

    @abstractmethod
    def rewrite_query(self, query: str) -> str:
        """
        Rewrite a query to disambiguate it and make it more self-contained.

        For example, the input

            John: People differ in their appreciation of ice cream.
            Mary: Do you like it?

        can be rewritten as

            Does John like ice cream?

        :param query: Query preceded by some context
        :return: Rewritten query
        """

        pass

    @abstractmethod
    def rewrite_memory(self, passage: str, context: str) -> str:
        """
        Rewrite a passage to disambiguate it and make it more self-contained.

        For example, the passage "Mary: I like it!" together with the context
        "People differ in their appreciation of ice cream." can be rewritten as "Mary likes ice cream".

        :param passage: The memory to be rewritten
        :param context: Context that helps disambiguate the memory
        :return: Rewritten memory
        """

        pass


class OpenAIRewriteModel(BaseRewriteModel):
    """
    Rewrite model wrapper for OpenAI completion models such as text-davinci-003.
    """

    def __init__(self,
                 model_name: str = "text-davinci-003",
                 api_key: str = None,
                 temperature: float = 0.7,
                 max_tokens: int = 512,
                 query_rewrite_prompt: str = None,
                 memory_rewrite_prompt: str = None,
                 max_input_chars: int = 2048):
        self.model_name = model_name
        if api_key:
            self.api_key = api_key
        elif 'OPENAI_API_KEY' in os.environ:
            self.api_key = os.environ['OPENAI_API_KEY']
        else:
            raise ValueError("An OpenAI api key needs to be provided as a parameter or environment variable")
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.query_prompt_template = query_rewrite_prompt or DEFAULT_QUERY_PROMPT_TEMPLATE
        self.memory_prompt_template = memory_rewrite_prompt or DEFAULT_MEMORY_PROMPT_TEMPLATE
        self.max_input_chars = max_input_chars

    def rewrite_query(self, query: str) -> str:
        self.check_inputs([query])
        prompt = self.make_query_prompt(query)
        return self.post_process_query(self.completion(prompt))

    def rewrite_memory(self, passage: str, context: str) -> str:
        self.check_inputs([passage, context])
        prompt = self.make_memory_prompt(passage, context)
        return self.post_process_memory(self.completion(prompt))

    def completion(self, prompt) -> str:
        openai.api_key = self.api_key
        response = openai.Completion.create(model=self.model_name, prompt=prompt, temperature=self.temperature,
                                            max_tokens=self.max_tokens)
        return response['choices'][0]['text'].strip()

    def make_query_prompt(self, query) -> str:
        return string.Template(self.query_prompt_template).substitute(query=query)

    def make_memory_prompt(self, passage, context) -> str:
        return string.Template(self.memory_prompt_template).substitute(passage=passage, context=context)

    def post_process_query(self, query: str) -> str:
        return json.loads(query)["processed"]

    def post_process_memory(self, memory: str) -> str:
        return "\n".join(json.loads(memory)["processed"].values())

    def check_inputs(self, inputs: List[str]):
        if sum([len(e) for e in inputs]) >= self.max_input_chars:
            raise ValueError(f"Input length exceeding {self.max_input_chars} characters")


class OpenAIChatRewriteModel(OpenAIRewriteModel):
    """
    Rewrite model wrapper for OpenAI chat models such as gpt-3.5-turbo and gpt-4.
    """

    def __init__(self,
                 model_name: str = "gpt-3.5-turbo",
                 api_key: str = None,
                 temperature: float = 0.7,
                 max_tokens: int = 512,
                 query_rewrite_prompt: str = None,
                 memory_rewrite_prompt: str = None,
                 query_rewrite_system_message: str = None,
                 memory_rewrite_system_message: str = None,
                 max_input_chars: int = 2048):
        super().__init__(model_name, api_key, temperature, max_tokens, query_rewrite_prompt, memory_rewrite_prompt,
                         max_input_chars)
        self.query_rewrite_system_message = query_rewrite_system_message or DEFAULT_QUERY_SYSTEM_MESSAGE
        self.memory_rewrite_system_message = memory_rewrite_system_message or DEFAULT_MEMORY_SYSTEM_MESSAGE

    def rewrite_query(self, query: str) -> str:
        self.check_inputs([query])
        prompt = self.make_query_prompt(query)
        return self.post_process_query(self.chat_completion(prompt, self.query_rewrite_system_message))

    def rewrite_memory(self, passage: str, context: str) -> str:
        self.check_inputs([passage, context])
        prompt = self.make_memory_prompt(passage, context)
        return self.post_process_memory(self.chat_completion(prompt, self.memory_rewrite_system_message))

    def chat_completion(self, prompt, system_message) -> str:
        messages = [
            {"role": "system", "content": system_message},
            {"role": "user", "content": prompt}
        ]

        openai.api_key = self.api_key
        response = openai.ChatCompletion.create(model=self.model_name, messages=messages, temperature=self.temperature,
                                                max_tokens=self.max_tokens)
        return response['choices'][0]['message']['content'].strip()



import unittest
from goodai.ltm.mem.chunk_queue import ChunkQueue, ChunkExpansionOptions


class TestChunkQueue(unittest.TestCase):
    def test_simple_insertion(self):
        seq_len = 8
        queue_capacity = 10
        num_seqs_in_chunk = 3
        chunk_capacity = seq_len * num_seqs_in_chunk
        chunk_index_at_overlap = chunk_capacity // 2
        self.assertEqual(12, chunk_index_at_overlap)
        sq = ChunkQueue(queue_capacity, chunk_capacity, chunk_index_at_overlap)
        start = 0
        token_ids = list(range(start, start + seq_len))
        sq.add_sequence(token_ids, None)
        token_ids = list(range(start + seq_len, start + seq_len + seq_len))
        sq.add_sequence(token_ids, None)
        self.assertEqual(2, len(sq.chunks))
        self.assertEqual(16, len(sq.token_ids))

    def test_insertion(self):
        seq_len = 8
        queue_capacity = 10
        num_seqs_in_chunk = 3
        chunk_capacity = seq_len * num_seqs_in_chunk
        chunk_index_at_overlap = chunk_capacity // 2
        sq = ChunkQueue(queue_capacity, chunk_capacity, chunk_index_at_overlap)
        for i in range(queue_capacity * 3):
            for j in range(num_seqs_in_chunk):
                start = i * chunk_capacity + j * seq_len
                token_ids = list(range(start, start + seq_len))
                sq.add_sequence(token_ids, None)
        assert sq.get_queue_size() == queue_capacity
        last_token_id = queue_capacity * 3 * chunk_capacity - 1
        sqs = sq.get_queue_size()
        stored_token_ids = []
        for i in range(0, sqs, 2):
            chunk = sq.chunks[i]
            observed_chunk = sq.get_chunk_token_ids(chunk)
            stored_token_ids.extend(observed_chunk)
        assert stored_token_ids[-1] == last_token_id
        for i in range(1, len(stored_token_ids)):
            assert stored_token_ids[i] == stored_token_ids[i-1] + 1
        stored_token_ids = []
        for i in range(1, sqs, 2):
            chunk = sq.chunks[i]
            observed_chunk = sq.get_chunk_token_ids(chunk)
            stored_token_ids.extend(observed_chunk)
        for i in range(1, len(stored_token_ids)):
            assert stored_token_ids[i] == stored_token_ids[i-1] + 1

    def test_full_capacity_insertion(self):
        queue_capacity = 10
        chunk_capacity = 24
        chunk_index_at_overlap = chunk_capacity // 2
        sq = ChunkQueue(queue_capacity, chunk_capacity, chunk_index_at_overlap)
        sequence = list(range(chunk_capacity * 3))
        sq.add_sequence(sequence, None)
        sqs = sq.get_queue_size()
        stored_token_ids = []
        for i in range(0, sqs, 2):
            chunk = sq.chunks[i]
            observed_chunk = sq.get_chunk_token_ids(chunk)
            stored_token_ids.extend(observed_chunk)
        assert stored_token_ids == sequence

    def test_retrieve_complete_sequences_adds_tokens_from_adjacent_chunks(self):
        chunk_capacity = 5
        chunk_index_at_overlap = chunk_capacity // 2
        _chunk_queue = ChunkQueue(12, chunk_capacity, chunk_index_at_overlap)
        _chunk_queue.add_sequence([1, 2, 3], None)
        _chunk_queue.add_sequence([4, 5, 6], None)
        _chunk_queue.add_sequence([7, 8, 9], None)
        _chunk_queue.add_sequence([10, 11, 12], None)
        _chunk_queue.add_sequence([10, 8, 12], None)
        _chunk_queue.add_sequence([14, 11, 12], None)
        _chunk_queue.add_sequence([3, 8, 21], None)
        _chunk_queue.add_sequence([13, 12, 9], None)
        _chunk_queue.add_sequence([3, 4, 28], None)
        _chunk_queue.add_sequence([12, 8, 25], None)

        punctuation_ids = {3, 8}
        ce_options = ChunkExpansionOptions.default(chunk_capacity, punctuation_ids)

        chunk_ids = [3, 5]
        result = _chunk_queue.retrieve_complete_sequences(chunk_ids, ce_options)

        self.assertEqual(2, len(result)), f'got {len(result)} sequences'
        seq1 = result[0]
        seq2 = result[1]

        self.assertEqual([7, 8, 9, 10, 11, 12, 10, 8], seq1)
        self.assertEqual([9, 10, 11, 12, 10, 8, 12, 14, 11, 12, 3], seq2)

    def test_zero_chunk_overlap(self):
        chunk_capacity = 5
        chunk_index_at_overlap = chunk_capacity
        _chunk_queue = ChunkQueue(25, chunk_capacity, chunk_index_at_overlap, first_token_seq_id=17)
        _chunk_queue.add_sequence([1, 2, 3], None)
        _chunk_queue.add_sequence([4, 5, 6], None)
        _chunk_queue.add_sequence([7, 8, 9], None)
        _chunk_queue.add_sequence([10, 11, 12], None)
        _chunk_queue.add_sequence([10, 8, 12], None)
        _chunk_queue.add_sequence([14, 11, 12], None)
        _chunk_queue.add_sequence([3, 8, 21], None)
        _chunk_queue.add_sequence([13, 12, 9], None)
        _chunk_queue.add_sequence([3, 4, 28], None)
        _chunk_queue.add_sequence([12, 8, 25], None)

        punctuation_ids = {3, 8}
        ce_options = ChunkExpansionOptions.default(chunk_capacity, punctuation_ids)

        chunk_ids = [3, 5]
        result = _chunk_queue.retrieve_complete_sequences(chunk_ids, ce_options)

        self.assertEqual(2, len(result)), f'got {len(result)} sequences'
        seq1 = result[0]
        seq2 = result[1]

        self.assertEqual([12, 14, 11, 12, 3, 8, 21, 13, 12, 9, 3], seq1)
        self.assertEqual([4, 28, 12, 8, 25], seq2)

    def test_overlap_of_1_token(self):
        chunk_capacity = 5
        chunk_index_at_overlap = chunk_capacity - 1
        _chunk_queue = ChunkQueue(25, chunk_capacity, chunk_index_at_overlap, first_token_seq_id=39)
        _chunk_queue.add_sequence([1, 2, 3], None)
        _chunk_queue.add_sequence([4, 5, 6], None)
        _chunk_queue.add_sequence([], None)
        _chunk_queue.add_sequence([7, 8, 9], None)
        _chunk_queue.add_sequence([10, 11, 12], None)
        _chunk_queue.add_sequence([10, 8, 12], None)
        _chunk_queue.add_sequence([], None)
        _chunk_queue.add_sequence([14, 11, 12], None)
        _chunk_queue.add_sequence([3, 8, 21], None)
        _chunk_queue.add_sequence([13, 12, 9], None)
        _chunk_queue.add_sequence([3, 4, 28], None)
        _chunk_queue.add_sequence([12, 8, 25], None)

        punctuation_ids = {3, 8}
        ce_options = ChunkExpansionOptions.default(chunk_capacity, punctuation_ids)

        chunk_ids = [3, 5]
        result = _chunk_queue.retrieve_complete_sequences(chunk_ids, ce_options)

        self.assertEqual(2, len(result)), f'got {len(result)} sequences'
        seq1 = result[0]
        seq2 = result[1]

        self.assertEqual([9, 10, 11, 12, 10, 8, 12, 14, 11, 12, 3], seq1)
        self.assertEqual([21, 13, 12, 9, 3, 4, 28, 12, 8], seq2)

    def test_separators(self):
        chunk_capacity = 5
        chunk_index_at_overlap = chunk_capacity // 2
        _chunk_queue = ChunkQueue(25, chunk_capacity, chunk_index_at_overlap, first_token_seq_id=75)
        _chunk_queue.add_sequence([1, 2, 3], None)
        _chunk_queue.add_sequence([4, 5, 6], None)
        _chunk_queue.add_separator()
        _chunk_queue.add_sequence([7, 8, 9], None)
        _chunk_queue.add_sequence([10, 11, 12], None)
        _chunk_queue.add_separator()
        _chunk_queue.add_sequence([10, 8, 12], None)
        _chunk_queue.add_sequence([14, 11, 12], None)
        _chunk_queue.add_separator()
        _chunk_queue.add_sequence([3, 8, 21], None)
        _chunk_queue.add_sequence([13, 12, 9], None)
        _chunk_queue.add_separator()
        _chunk_queue.add_sequence([3, 4, 28], None)
        _chunk_queue.add_sequence([12, 8, 25], None)

        punctuation_ids = {3, 8}
        ce_options = ChunkExpansionOptions.default(chunk_capacity, punctuation_ids)

        chunk_ids = [3, 6, 9]
        result = _chunk_queue.retrieve_complete_sequences(chunk_ids, ce_options)

        self.assertEqual(3, len(result)), f'got {len(result)} sequences'
        expected = [[7, 8, 9, 10, 11, 12], [10, 8, 12, 14, 11, 12], [3, 8, 21, 13, 12, 9]]
        for e_seq, r_seq in zip(expected, result):
            self.assertEqual(e_seq, r_seq)

    def test_overflow_with_separators(self):
        chunk_capacity = 5
        queue_capacity = 10
        chunk_index_at_overlap = chunk_capacity // 2
        _chunk_queue = ChunkQueue(queue_capacity, chunk_capacity, chunk_index_at_overlap, first_token_seq_id=75)
        for i in range(200):
            _chunk_queue.add_sequence([1, 2, 3], None)
            _chunk_queue.add_separator()
        self.assertEqual(6, len(_chunk_queue.separator_seq_ids))
        self.assertTrue(_chunk_queue.separator_seq_ids[0] >= _chunk_queue.first_token_seq_id)
        self.assertTrue(_chunk_queue.separator_seq_ids[-1] <= _chunk_queue.first_token_seq_id +
                        len(_chunk_queue.token_ids))

    def test_flush(self):
        chunk_capacity = 5
        queue_capacity = 10
        chunk_index_at_overlap = chunk_capacity // 2
        _chunk_queue = ChunkQueue(queue_capacity, chunk_capacity, chunk_index_at_overlap, first_token_seq_id=220)
        for i in range(100):
            _chunk_queue.add_sequence([1, 2, 3], None)
            _chunk_queue.add_separator()
        _chunk_queue.flush()
        self.assertTrue(len(_chunk_queue.separator_seq_ids) == 0)
        self.assertTrue(len(_chunk_queue.chunks) == 0)
        self.assertTrue(len(_chunk_queue.token_ids) == 0)
        self.assertTrue(len(_chunk_queue.chunk_map) == 0)
        self.assertTrue(_chunk_queue.first_token_seq_id == 0)

    def test_replacement_1(self):
        chunk_capacity = 5
        chunk_index_at_overlap = chunk_capacity // 2
        _chunk_queue = ChunkQueue(25, chunk_capacity, chunk_index_at_overlap, first_token_seq_id=73)
        _, k1 = _chunk_queue.add_sequence([1, 2, 3], None)
        _, k2 = _chunk_queue.add_sequence([4, 5, 6], None)
        _, k3 = _chunk_queue.add_sequence([7, 8, 9], None)

        chunk_sequences = _chunk_queue.get_chunk_sequences()
        expected_cs_1 = [[1, 2, 3, 4, 5], [3, 4, 5, 6, 7], [5, 6, 7, 8, 9], [7, 8, 9], [9]]
        self.assertEqual(expected_cs_1, chunk_sequences)

        _chunk_queue.replace_sequence(k2, [8, 9, 10, 9])
        chunk_sequences = _chunk_queue.get_chunk_sequences()
        expected_cs_2 = [[1, 2, 3, 8, 9], [3, 8, 9, 10, 9], [9, 10, 9], [9], [7, 8, 9], [9]]
        self.assertEqual(expected_cs_2, chunk_sequences)

    def test_replacement_2(self):
        chunk_capacity = 5
        chunk_index_at_overlap = chunk_capacity // 2
        _chunk_queue = ChunkQueue(25, chunk_capacity, chunk_index_at_overlap, first_token_seq_id=73)
        _chunk_queue.add_separator()
        _, k1 = _chunk_queue.add_sequence([1, 2, 3], None)
        _, k2 = _chunk_queue.add_sequence([4, 5, 6], None)
        _, k3 = _chunk_queue.add_sequence([7, 8, 9], None)

        chunk_sequences = _chunk_queue.get_chunk_sequences()
        expected_cs_1 = [[1, 2, 3, 4, 5], [3, 4, 5, 6, 7], [5, 6, 7, 8, 9], [7, 8, 9], [9]]
        self.assertEqual(expected_cs_1, chunk_sequences)

        _chunk_queue.replace_sequence(k1, [])
        chunk_sequences = _chunk_queue.get_chunk_sequences()
        expected_cs_2 = [[4], [5, 6, 7, 8, 9], [7, 8, 9], [9]]
        self.assertEqual(expected_cs_2, chunk_sequences)

    def test_replacement_3(self):
        chunk_capacity = 5
        chunk_index_at_overlap = chunk_capacity // 2
        _chunk_queue = ChunkQueue(25, chunk_capacity, chunk_index_at_overlap, first_token_seq_id=43)
        _, k1 = _chunk_queue.add_sequence([1, 2, 3], None)
        _, k2 = _chunk_queue.add_sequence([4, 5, 6], None)
        _chunk_queue.add_separator()
        _, k3 = _chunk_queue.add_sequence([7, 8, 9], None)

        chunk_sequences = _chunk_queue.get_chunk_sequences()
        expected_cs_1 = [[1, 2, 3, 4, 5], [3, 4, 5, 6], [5, 6], [7, 8, 9], [9]]
        self.assertEqual(expected_cs_1, chunk_sequences)

        _chunk_queue.replace_sequence(k3, [11, 12, 11, 12, 11])

        chunk_sequences = _chunk_queue.get_chunk_sequences()
        expected_cs_2 = [[1, 2, 3, 4, 5],
                         [3, 4, 5, 6],
                         [5, 6],
                         [11, 12, 11, 12, 11],
                         [11, 12, 11],
                         [11]]
        self.assertEqual(expected_cs_2, chunk_sequences)

        chunk_ids = [_chunk_queue.chunks[-4].chunk_id, _chunk_queue.chunks[-1].chunk_id]
        punctuation_ids = {90}
        ce_options = ChunkExpansionOptions.default(chunk_capacity, punctuation_ids)
        result = _chunk_queue.retrieve_complete_sequences(chunk_ids, ce_options)
        expected_result = [[1, 2, 3, 4, 5, 6], [11, 12, 11, 12, 11]]
        self.assertEqual(expected_result, result)

    def test_replacement_4(self):
        chunk_capacity = 5
        chunk_index_at_overlap = chunk_capacity // 2
        _chunk_queue = ChunkQueue(25, chunk_capacity, chunk_index_at_overlap, first_token_seq_id=43)
        _, k1 = _chunk_queue.add_sequence([1, 2, 3], None)
        _, k2 = _chunk_queue.add_sequence([4, 5, 6], None)
        _chunk_queue.add_separator()
        _, k3 = _chunk_queue.add_sequence([7, 8, 9], None)

        chunk_sequences = _chunk_queue.get_chunk_sequences()
        expected_cs_1 = [[1, 2, 3, 4, 5], [3, 4, 5, 6], [5, 6], [7, 8, 9], [9]]
        self.assertEqual(expected_cs_1, chunk_sequences)

        _chunk_queue.replace_sequence(k2, [11, 12, 13, 14, 15, 11, 12, 13, 14, 15])
        chunk_sequences = _chunk_queue.get_chunk_sequences()
        expected_cs_2 = [[1, 2, 3, 11, 12],
                         [3, 11, 12, 13, 14],
                         [12, 13, 14, 15, 11],
                         [14, 15, 11, 12, 13],
                         [11, 12, 13, 14, 15],
                         [13, 14, 15],
                         [15],
                         [7, 8, 9],
                         [9]]
        self.assertEqual(expected_cs_2, chunk_sequences)

        chunk_ids = [_chunk_queue.chunks[-4].chunk_id]
        ce_options = ChunkExpansionOptions(minSideTokens=0, maxSideTokens=100, leftStopAfterTokenIds=[],
                                           rightStopAtTokenIds=[])
        result = _chunk_queue.retrieve_complete_sequences(chunk_ids, ce_options)
        expected_result = [[1, 2, 3, 11, 12, 13, 14, 15, 11, 12, 13, 14, 15]]
        self.assertEqual(expected_result, result)



from dataclasses import dataclass
from typing import List, Optional, Tuple
from transformers import PreTrainedTokenizer

from goodai.helpers.tokenizer_helper import get_token_index


@dataclass
class QATokenizedEntry:
    ds_name: str
    e_id: str
    context: str
    answer_token_ids: List[int]
    answer_seq_index: int
    question: str
    content_token_ids: Optional[List[int]] = None
    question_token_ids: Optional[List[int]] = None

    @staticmethod
    def from_example(example: dict, tokenizer: PreTrainedTokenizer) -> 'QATokenizedEntry':
        ds_name = example['ds_name']
        e_id = example['id']
        context = example.get('context')
        if context is None:
            context = example.get('story')
        question = example.get('question')
        question_ids = tokenizer.encode(' ' + question, add_special_tokens=False)
        answers = example['answers']
        answer_text_list = answers['text']
        answer_text = answer_text_list[0]
        answer_index = answers['answer_start'][0]
        tok_output = tokenizer(context, return_offsets_mapping=True, add_special_tokens=False)
        context_ids = tok_output['input_ids']
        offset_mapping: List[Tuple[int, int]] = tok_output['offset_mapping']
        answer_token_index = get_token_index(context_ids, offset_mapping, answer_index)
        answer_token_ids = tokenizer.encode(' ' + answer_text, add_special_tokens=False)
        return QATokenizedEntry(ds_name=ds_name, e_id=e_id, context=context,
                                content_token_ids=context_ids,
                                answer_token_ids=answer_token_ids,
                                answer_seq_index=answer_token_index,
                                question_token_ids=question_ids,
                                question=question)



from dataclasses import dataclass
from typing import List


@dataclass
class QueryPassageExample:
    queryIds: List[int]
    """
    Query token IDs
    """

    passageIds: List[int]
    """
    Passage token IDs
    """

    match: bool
    """
    Whether the query and the passage are a good match
    """



from dataclasses import dataclass
from typing import Optional, Tuple

import numpy as np


@dataclass
class SimpleVectorDb:
    """
    Simple vector database implementation.
    """
    all_vectors: Optional[np.ndarray] = None
    all_ids: Optional[np.ndarray] = None

    def search(self, vectors: np.ndarray, k: int = 1, max_batch_size=256) -> Tuple[np.ndarray, np.ndarray]:
        dist_list = []
        idx_list = []
        for b0 in range(0, vectors.shape[0], max_batch_size):
            b_vectors = vectors[b0:b0 + max_batch_size]
            b_dist, b_idx = self._search_direct(b_vectors, k=k)
            dist_list.append(b_dist)
            idx_list.append(b_idx)
        return np.concatenate(dist_list), np.concatenate(idx_list),

    def _search_direct(self, vectors: np.ndarray, k: int = 1) -> Tuple[np.ndarray, np.ndarray]:
        batch_size = vectors.shape[0]
        placeholder_dist = np.ones((batch_size, k)) * 1e+38
        placeholder_ids = -np.ones((batch_size, k), dtype=np.int64)
        if self.all_vectors is None:
            return placeholder_dist, placeholder_ids,
        # vectors: (n, emb_size,)
        # all_vectors: (m, emb_size,)
        diff_sq = (vectors[:, None, :] - self.all_vectors[None, :, :]) ** 2
        # diff_sq: (n, m, emb_size,)
        all_sq_distances = np.sum(diff_sq, axis=2)
        # mean_diff_sq: (n, m,)
        sort_indexes = np.argsort(all_sq_distances, axis=1)
        # sort_indexes: (n, m,)
        select_indexes = sort_indexes[:, :k]
        flat_select_indexes = select_indexes.flatten()
        flat_result_ids = self.all_ids[flat_select_indexes]
        # result_indexes: (n, k,)
        result_size, num_result_ids = select_indexes.shape
        result_range = np.arange(0, result_size)
        rep_range = np.repeat(result_range, num_result_ids)
        result_distances = all_sq_distances[rep_range, flat_select_indexes]
        result_distances = np.reshape(result_distances, select_indexes.shape)
        result_ids = np.reshape(flat_result_ids, select_indexes.shape)
        placeholder_dist[:, :result_distances.shape[1]] = result_distances
        placeholder_ids[:, :result_ids.shape[1]] = result_ids
        return placeholder_dist, placeholder_ids,

    def add_with_ids(self, vectors: np.ndarray, ids: np.ndarray):
        if vectors.shape[0] != ids.shape[0]:
            raise ValueError('Mismatch in number of vectors and ids provided')
        self.remove_ids(ids)
        assert len(ids.shape) == 1, 'ids must be 1-dimensional'
        if self.all_vectors is None:
            self.all_vectors = vectors
        else:
            self.all_vectors = np.concatenate([self.all_vectors, vectors], axis=0)
        if self.all_ids is None:
            self.all_ids = ids
        else:
            self.all_ids = np.concatenate([self.all_ids, ids], axis=0)

    def reset(self):
        self.all_vectors = None
        self.all_ids = None

    def remove_ids(self, ids: np.ndarray):
        if len(ids.shape) != 1:
            raise ValueError('ids must be 1-dimensional')
        to_remove = np.isin(self.all_ids, ids)
        if np.any(to_remove):
            if np.all(to_remove):
                self.reset()
            else:
                keep = ~to_remove
                self.all_ids = self.all_ids[keep]
                self.all_vectors = self.all_vectors[keep, :]

    def __eq__(self, other):
        if not isinstance(other, SimpleVectorDb):
            return False
        return np.array_equal(self.all_vectors, other.all_vectors) and np.array_equal(self.all_ids, other.all_ids)



import abc
import logging
from abc import ABC
from dataclasses import dataclass
from typing import List, Dict, Tuple

import numpy as np
from tqdm import tqdm
from transformers import PreTrainedTokenizer

from goodai.ltm.data.names import NameSource
from goodai.ltm.eval.metrics import get_correctness_score
from goodai.ltm.mem.base import BaseTextMemory


@dataclass
class QAScenario:
    context: List[str]
    supportingFacts: List[str]
    question: str


class BaseMemEvaluator(ABC):
    """
    Abstract base class for text memory model evaluators.
    """

    def __init__(self, tokenizer: PreTrainedTokenizer, top_ks: List[int],
                 max_query_tokens: int, has_query_noise: bool,
                 add_names_to_context: bool = True, correctness_threshold=70):
        if len(top_ks) == 0:
            raise ValueError('At least one top-k must be provided')
        self.top_ks = top_ks
        self.correctness_threshold = correctness_threshold
        self.tokenizer = tokenizer
        self.max_query_tokens = max_query_tokens
        self.has_query_noise = has_query_noise
        self.add_names_to_context = add_names_to_context
        self.rnd1 = np.random.RandomState(8001)
        self.rnd2 = np.random.RandomState(8002)

    @abc.abstractmethod
    def get_facts_to_be_stored(self) -> List[str]:
        pass

    @abc.abstractmethod
    def get_scenarios(self) -> List[QAScenario]:
        pass

    def get_query(self, scenario: QAScenario) -> str:
        if self.add_names_to_context:
            name1, name2 = tuple(NameSource.get_instance().sample_first_names(self.rnd1, count=2))
            names = [name1, name2]
            names_context = [f'{names[i % 2]}: {ctx}' for i, ctx in enumerate(scenario.context)]
            c_len = len(names_context)
            q_name = names[c_len % 2]
            a_name = names[(c_len + 1) % 2]
            query = f'{q_name}: {scenario.question}\n{a_name}:'
        else:
            names_context = scenario.context
            query = scenario.question
        truncate_q_ids_at = self.max_query_tokens
        if self.has_query_noise:
            no_noise_ids = self.tokenizer.encode(query, add_special_tokens=False)
            len_nni = len(no_noise_ids)
            if len_nni < self.max_query_tokens:
                truncate_q_ids_at = self.rnd2.randint(max(1, len_nni), self.max_query_tokens + 1)
            context_as_text = '\n'.join(names_context[-3:])
            query = context_as_text + '\n' + query
        query_token_ids = self.tokenizer.encode(query, add_special_tokens=False)
        query_token_ids = query_token_ids[-truncate_q_ids_at:]
        return self.tokenizer.decode(query_token_ids, skip_special_tokens=True)

    def get_queries_and_support(self, scenarios: List[QAScenario]) -> Tuple[List[str], List[List[str]]]:
        queries = []
        supports_list: List[List[str]] = []
        for scenario in scenarios:
            query = self.get_query(scenario)
            support = scenario.supportingFacts
            queries.append(query)
            supports_list.append(support)
        return queries, supports_list,

    def cross_max_correctness(self, retrieved_texts: List[str], supporting_facts: List[str]):
        result = []
        for rt in retrieved_texts:
            scores = [get_correctness_score(self.tokenizer, rt, f) for f in supporting_facts]
            result.append(max(scores))
        return result

    def evaluate(self, memory: BaseTextMemory) -> Dict[str, float]:
        facts = self.get_facts_to_be_stored()
        facts_text = '\n'.join(facts)
        memory.add_text(facts_text + '\n', show_progress_bar=True)
        scenarios = self.get_scenarios()
        queries, supports = self.get_queries_and_support(scenarios)
        k = max(self.top_ks)
        retrieved = memory.retrieve_multiple(queries, k=k, show_progress_bar=True)
        top_k_map: Dict[int, int] = dict()
        for q, s_retrieved, s_supports in tqdm(zip(queries, retrieved, supports), desc='Comparison', unit='scenario'):
            s_retrieved_texts = [r.passage for r in s_retrieved]
            if len(s_retrieved_texts) == 0:
                logging.warning(f'No memories retrieved with query "{q}"')
                continue
            correctness_values = self.cross_max_correctness(s_retrieved_texts, s_supports)
            for top_k in self.top_ks:
                selected_cv = correctness_values[:top_k]
                if max(selected_cv) >= self.correctness_threshold:
                    top_k_map[top_k] = top_k_map.get(top_k, 0) + 1
        item_count = len(retrieved)
        return {f'ACC@{top_k}': v / item_count for top_k, v in top_k_map.items()}



from typing import List, Optional

from transformers import PreTrainedTokenizer

from goodai.ltm.eval.msmarco import MsMarcoMemEvaluator
from goodai.ltm.eval.qp_ds import QPDSMemEvaluator
from goodai.ltm.eval.qrecc import QreccMemEvaluator
from goodai.ltm.eval.strategy_qa import StrategyQAMemEvaluator


class AutoMemEvaluator:
    """
    Factory class for memory evaluators.
    """

    @staticmethod
    def create(name: str, tokenizer: PreTrainedTokenizer, top_ks: List[int], max_query_tokens: int,
               has_query_noise: bool, max_scenarios: Optional[int] = None):
        if max_scenarios is None:
            max_scenarios = 2000
        if name == 'qrecc':
            return QreccMemEvaluator(tokenizer, top_ks, max_query_tokens, has_query_noise,
                                     max_scenarios=max_scenarios)
        elif name == 'strategyqa':
            return StrategyQAMemEvaluator(tokenizer, top_ks, max_query_tokens, has_query_noise,
                                          max_scenarios=max_scenarios)
        elif name == 'msmarco':
            return MsMarcoMemEvaluator(tokenizer, top_ks, max_query_tokens, has_query_noise,
                                       max_scenarios=max_scenarios)
        elif name in ['qp_squad_v2', 'qp_coqa', 'qp_wiki', 'qp_wikianswers']:
            return QPDSMemEvaluator(name[3:], tokenizer, top_ks, max_query_tokens, has_query_noise,
                                    max_scenarios=max_scenarios)
        else:
            raise ValueError(f'Dataset name "{name}" not recognized')



from abc import ABC, abstractmethod
from typing import List, Union

import numpy as np
import torch


class BaseTextEmbeddingModel(ABC):
    """
    Abstract base class for text embedding models.

    The interface allows different embeddings for queries (retrieval embeddings) and passages
    (storage embeddings) as well as multiple retrieval and storage embeddings for a query or passage.
    """

    @abstractmethod
    def get_embedding_dim(self) -> int:
        """
        :return: The number of dimensions of embedding vectors.
        """
        pass

    @abstractmethod
    def get_num_retrieval_embeddings(self) -> int:
        """
        :return: The number of retrieval embeddings produced by the model, per query.
        """
        pass

    @abstractmethod
    def get_num_storage_embeddings(self) -> int:
        """
        :return: The number of storage embeddings produced by the model, per passage.
        """
        pass

    @abstractmethod
    def get_info(self) -> str:
        """
        :return: Information about the model.
        """
        pass

    @abstractmethod
    def encode_queries(self, queries: List[str], batch_size: int = 64, show_progress_bar: bool = False,
                       convert_to_tensor: bool = False,
                       device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        pass

    @abstractmethod
    def encode_corpus(self, passages: List[str], batch_size: int = 64, show_progress_bar: bool = False,
                      convert_to_tensor: bool = False,
                      device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        pass



from typing import Iterator, List, Union
from torch import nn, optim
from torch.optim.lr_scheduler import LambdaLR
from transformers import get_scheduler


class ScheduledOptimizer:
    def __init__(self, parameters: Union[Iterator[nn.Parameter], List[dict]],
                 lr: float, num_training_steps: int, num_warmup_steps: int = 0,
                 weight_decay: float = 1e-3):
        super().__init__()
        self.opt = optim.AdamW(parameters, lr=lr, weight_decay=weight_decay)
        self.scheduler = self.get_opt_scheduler(self.opt, num_training_steps, num_warmup_steps=num_warmup_steps)

    @staticmethod
    def get_opt_scheduler(optimizer: optim.Optimizer, num_training_steps: int, num_warmup_steps=0):
        return get_scheduler(
            'linear',
            optimizer=optimizer,
            num_warmup_steps=num_warmup_steps,
            num_training_steps=num_training_steps,
        )

    def zero_grad(self, set_to_none: bool = False):
        self.opt.zero_grad(set_to_none=set_to_none)

    def step(self):
        self.opt.step()
        self.scheduler.step()

    @staticmethod
    def _get_last_lr(scheduler:  LambdaLR):
        last_lr_list = scheduler.get_last_lr()
        return last_lr_list[0] if len(last_lr_list) > 0 else 0

    def get_last_lr(self):
        return self._get_last_lr(self.scheduler)

    def get_last_lrs(self) -> List[float]:
        return self.scheduler.get_last_lr()



import itertools
import logging
import math
import torch
from torch import nn
import torch.nn.functional as F
from tqdm import tqdm
from transformers import AutoModel, AutoTokenizer, PreTrainedTokenizer
from typing import List, Optional, Tuple

from goodai.helpers.tokenizer_helper import get_attention_after_token, get_model_inputs
from goodai.helpers.torch_helper import param_count
from goodai.ltm.embeddings.contrast_classifier import ContrastClassifier
from goodai.ltm.reranking.base import BaseTextMatchingModel
from goodai.ltm.reranking.prob_model import BaseQueryPassageProbModel


class DefaultRerankingCrossEncoder(nn.Module, BaseQueryPassageProbModel, BaseTextMatchingModel):
    sep_id_tensor: torch.Tensor

    def __init__(self, model_name: str, default_query_seq_len: Optional[int] = None,
                 default_passage_seq_len: Optional[int] = None, num_end_chars_lb_ignore=18,
                 dropout=0.01):
        super().__init__()
        self.num_end_chars_lb_ignore = num_end_chars_lb_ignore
        self.default_passage_seq_len = default_passage_seq_len
        self.default_query_seq_len = default_query_seq_len
        self.model = AutoModel.from_pretrained(model_name, output_hidden_states=True)
        self.tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(model_name)
        hidden_size = self.model.config.hidden_size
        cls_emb_scale = math.sqrt(hidden_size)
        self.classifier = ContrastClassifier(cls_emb_scale)
        lb_token_ids = self.tokenizer.encode('\n', add_special_tokens=False)
        if len(lb_token_ids) != 1:
            logging.warning(f'Tokenizer {self.tokenizer.name_or_path} does not have a line break token.')
            self.lb_token_id = -1
        else:
            self.lb_token_id = lb_token_ids[0]
        if self.tokenizer.sep_token_id is None:
            self.tokenizer.sep_token_id = self.tokenizer.eos_token_id
        if self.tokenizer.sep_token_id is None:
            raise ValueError(f'Tokenizer {model_name} does not have SEP or EOS tokens!')
        if self.tokenizer.sep_token_id == self.tokenizer.pad_token_id:
            raise ValueError(f'Tokenizer {model_name} with SEP equal to PAD not supported!')
        sep_id_tensor = torch.as_tensor([[self.tokenizer.sep_token_id]], dtype=torch.long)
        self.register_buffer('sep_id_tensor', sep_id_tensor)
        self.dummy = nn.Parameter()
        self.passage_slfatt_model = nn.Sequential(
            nn.Dropout(p=dropout),
            nn.Linear(hidden_size, 5),
        )

    def get_info(self):
        return f'{self.__class__} | Params: {param_count(self) / 1e6:.4g} million'

    def get_device(self):
        return self.dummy.device

    def get_lm_parameters(self):
        return self.model.parameters()

    def get_added_parameters(self):
        return itertools.chain(self.passage_slfatt_model.parameters(),
                               self.classifier.parameters())

    @staticmethod
    def get_embedding(hidden_states: torch.Tensor, attention_mask: torch.Tensor,
                      att_model: Optional[nn.Module]) -> torch.Tensor:
        # hidden_states: (batch_size, seq_len, emb_size,)
        x_att_mask = attention_mask[:, :, None]
        if att_model:
            # x_att_mask: (batch_size, seq_len, 1,)
            slf_att_logits = att_model(hidden_states)
            # slf_att_logits: (batch_size, seq_len, total_keys,)
            slf_att_logits = slf_att_logits - 200.0 * (1.0 - x_att_mask)
            slf_att_logits = torch.clamp(slf_att_logits, min=-100, max=+80)
            slf_att_weights = torch.softmax(slf_att_logits, dim=1)
            # slf_att_weights: (batch_size, seq_len, total_keys,)
            dot_product = hidden_states[:, :, None, :] * slf_att_weights[:, :, :, None]
        else:
            dot_product = hidden_states[:, :, None, :] * x_att_mask[:, :, :, None]
        raw_w_mean = torch.sum(dot_product, dim=1)
        # raw_w_mean: (batch_size, total_keys, emb_size,)
        return F.normalize(raw_w_mean, dim=-1)

    def forward(self, query_input_ids: torch.Tensor, query_attention_mask: torch.Tensor,
                query_token_lengths: torch.Tensor,
                passage_input_ids: torch.Tensor, passage_attention_mask: torch.Tensor) -> torch.Tensor:
        batch_size, query_len = query_input_ids.size(0), query_input_ids.size(1)
        passage_len = passage_input_ids.size(1)
        sep_t = torch.repeat_interleave(self.sep_id_tensor, batch_size, dim=0)
        input_ids = torch.hstack([query_input_ids, sep_t, passage_input_ids])
        attention_mask = torch.hstack([query_attention_mask, torch.ones_like(sep_t), passage_attention_mask])
        outputs = self.model(input_ids, attention_mask=attention_mask)
        last_hidden_states = outputs.last_hidden_state
        query_hidden_states = last_hidden_states[:, :query_len]
        passage_hidden_states = last_hidden_states[:, -passage_len:]
        lb_att_mask = get_attention_after_token(query_input_ids, query_attention_mask, query_token_lengths,
                                                self.lb_token_id, exclude_last_n_chars=self.num_end_chars_lb_ignore,
                                                device=self.get_device())
        emb_q_att_mask = query_attention_mask * lb_att_mask
        query_embedding = self.get_embedding(query_hidden_states, emb_q_att_mask, None)
        passage_embedding = self.get_embedding(passage_hidden_states, passage_attention_mask, self.passage_slfatt_model)
        match_probabilities = self.classifier(query_embedding, passage_embedding)
        return match_probabilities

    def get_probabilities_in_batches(self, query_input_ids: torch.Tensor, query_attention_mask: torch.Tensor,
                                     query_token_lengths: torch.Tensor,
                                     passage_input_ids: torch.Tensor, passage_attention_mask: torch.Tensor,
                                     max_batch_size: int, show_progress_bar: bool = False):
        batch_size = query_input_ids.size(0)
        output_probabilities_list = []
        rng = range(0, batch_size, max_batch_size)
        if show_progress_bar:
            rng = tqdm(rng, desc='Matching', unit='batch')
        for start_idx in rng:
            end_idx = start_idx + max_batch_size
            query_input_ids_batch = query_input_ids[start_idx:end_idx]
            query_attention_mask_batch = query_attention_mask[start_idx:end_idx]
            query_token_lengths_batch = query_token_lengths[start_idx:end_idx]
            passage_input_ids_batch = passage_input_ids[start_idx:end_idx]
            passage_attention_mask_batch = passage_attention_mask[start_idx:end_idx]
            output_probabilities_batch = self(query_input_ids_batch, query_attention_mask_batch,
                                              query_token_lengths_batch,
                                              passage_input_ids_batch, passage_attention_mask_batch)
            if not self.training:
                output_probabilities_batch = output_probabilities_batch.detach()
            output_probabilities_list.append(output_probabilities_batch)
        output_probabilities_t = torch.cat(output_probabilities_list)
        return output_probabilities_t

    def match_probabilities(self, sentences: List[Tuple[str, str]], fallback_max_query_tokens: int = 40,
                            fallback_max_passage_tokens: int = 70, show_progress_bar: bool = False,
                            max_batch_size: int = 50, use_preferred_seq_lengths: bool = True) -> List[float]:
        if len(sentences) == 0:
            return []
        max_query_tokens = None
        max_passage_tokens = None
        if use_preferred_seq_lengths:
            max_query_tokens = self.default_query_seq_len
            max_passage_tokens = self.default_passage_seq_len
        if max_query_tokens is None:
            max_query_tokens = fallback_max_query_tokens
        if max_passage_tokens is None:
            max_passage_tokens = fallback_max_passage_tokens
        tokenizer = self.tokenizer
        pad_id = tokenizer.pad_token_id
        query_ids_list = []
        passage_ids_list = []
        for query, passage in sentences:
            query_ids = tokenizer.encode(query, add_special_tokens=False)
            if len(query_ids) > max_query_tokens:
                query_ids = query_ids[-max_query_tokens:]
            passage_ids = tokenizer.encode(passage, add_special_tokens=False)
            if len(passage_ids) > max_passage_tokens:
                passage_ids = passage_ids[-max_passage_tokens:]
            passage_ids_list.append(passage_ids)
            query_ids_list.append(query_ids)
        device = self.get_device()
        query_seq_len = self.default_query_seq_len if use_preferred_seq_lengths else None
        passage_seq_len = self.default_passage_seq_len if use_preferred_seq_lengths else None
        query_inputs = get_model_inputs(query_ids_list, pad_id, device, prefix='query_',
                                        min_seq_len=query_seq_len, return_token_lengths=True,
                                        tokenizer=tokenizer)
        passage_inputs = get_model_inputs(passage_ids_list, pad_id, device, prefix='passage_',
                                          min_seq_len=passage_seq_len)
        probabilities_t = self.get_probabilities_in_batches(**query_inputs, **passage_inputs,
                                                            max_batch_size=max_batch_size,
                                                            show_progress_bar=show_progress_bar)
        # probabilities_t: (batch_size, 1,)
        return probabilities_t.squeeze(1).tolist()

    def predict(self, sentences: List[Tuple[str, str]], batch_size: int = 32,
                show_progress_bar: bool = False) -> List[float]:
        return self.match_probabilities(sentences, max_batch_size=batch_size, show_progress_bar=show_progress_bar)



import threading

import numpy as np
import openai
import torch
from typing import List, Union

from tqdm import tqdm

from goodai.ltm.embeddings.base import BaseTextEmbeddingModel

_openai_lock = threading.Lock()


class OpenAIEmbeddingModel(BaseTextEmbeddingModel):
    """
    Text embedding model based on OpenAI text embeddings.

    https://platform.openai.com/docs/guides/embeddings
    """

    def __init__(self, model_name: str = 'text-embedding-ada-002', emb_dim: int = 1536, api_key: str = None,
                 device: Union[torch.Tensor, str] = None):
        self.device = device
        self.api_key = api_key
        self.emb_dim = emb_dim
        self.model_name = model_name

    def get_embedding_dim(self) -> int:
        return self.emb_dim

    def get_num_retrieval_embeddings(self) -> int:
        return 1

    def get_num_storage_embeddings(self) -> int:
        return 1

    def get_info(self) -> str:
        return f'OpenAI embedding model "{self.model_name}" | Dimensions: {self.emb_dim}'

    def encode(self, sentences: List[str], batch_size: int = 64, show_progress_bar: bool = False,
               convert_to_tensor: bool = False,
               device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        if device is None:
            device = self.device
        with _openai_lock:
            if self.api_key is not None:
                openai.api_key = self.api_key
            rng = range(0, len(sentences), batch_size)
            if show_progress_bar:
                rng = tqdm(rng, desc='Embeddings', unit='batch')
            all_emb_vectors = []
            for b0 in rng:
                b_queries = sentences[b0:b0 + batch_size]
                response = openai.Embedding.create(input=b_queries, model=self.model_name)
                data_array = response['data']
                emb_vectors = [entry['embedding'] for entry in data_array]
                all_emb_vectors.extend(emb_vectors)
        if convert_to_tensor:
            return torch.as_tensor(all_emb_vectors, dtype=torch.float, device=device).unsqueeze(1)
        else:
            return np.array(all_emb_vectors)[:, None, :]

    def encode_queries(self, queries: List[str], batch_size: int = 64, show_progress_bar: bool = False,
                       convert_to_tensor: bool = False,
                       device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        return self.encode(queries, batch_size=batch_size, show_progress_bar=show_progress_bar,
                           convert_to_tensor=convert_to_tensor, device=device)

    def encode_corpus(self, passages: List[str], batch_size: int = 64, show_progress_bar: bool = False,
                      convert_to_tensor: bool = False,
                      device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        return self.encode(passages, batch_size=batch_size, show_progress_bar=show_progress_bar,
                           convert_to_tensor=convert_to_tensor, device=device)



import codecs
import json
import os
from typing import List, Optional

import datasets
import numpy as np
from transformers import PreTrainedTokenizer

from goodai.helpers.file_helper import download_zip
from goodai.ltm.eval.mem import BaseMemEvaluator, QAScenario


class MsMarcoMemEvaluator(BaseMemEvaluator):
    def __init__(self, tokenizer: PreTrainedTokenizer, top_ks: List[int],
                 max_query_tokens: int, has_query_noise: bool, use_rewrite: bool = True,
                 max_scenarios: int = 2000):
        super().__init__(tokenizer, top_ks, max_query_tokens, has_query_noise)
        self.use_rewrite = use_rewrite
        ds = datasets.load_dataset('ms_marco', 'v2.1')
        valid_data = ds['validation']
        self.data = valid_data[:max_scenarios]

    def get_facts_to_be_stored(self) -> List[str]:
        facts = []
        d = self.data
        answers, passages, query = d['answers'], d['passages'], d['query']
        for a, p, q in zip(answers, passages, query):
            is_selected = p['is_selected']
            if np.any(is_selected):
                a_len = [len(x) for x in a]
                a_idx = np.argmax(a_len)
                facts.append(a[a_idx])
        return facts

    @staticmethod
    def get_context(passages: dict) -> Optional[List[str]]:
        is_selected = passages['is_selected']
        if not np.any(is_selected):
            return None
        passage_text = passages['passage_text']
        return [p for p, s in zip(passage_text, is_selected) if not s]

    def get_scenarios(self) -> List[QAScenario]:
        scenarios = []
        d = self.data
        answers, passages, query = d['answers'], d['passages'], d['query']
        for a, p, q in zip(answers, passages, query):
            context = self.get_context(p)
            if context is not None:
                a_len = [len(x) for x in a]
                a_idx = np.argmax(a_len)
                sf = [a[a_idx]]
                scenario = QAScenario(context=context, supportingFacts=sf, question=q)
                scenarios.append(scenario)
        return scenarios



import unittest

from transformers import PreTrainedTokenizer, AutoTokenizer

from goodai.ltm.eval.metrics import get_correctness_score


class TestMetrics(unittest.TestCase):
    tokenizer: PreTrainedTokenizer

    def setUp(self) -> None:
        self.tokenizer = AutoTokenizer.from_pretrained('distilroberta-base')

    def test_exact_match_01(self):
        exp_t = "The sun is at the center of the solar system."
        pred_t = "the sun is at the center of the solar system"
        c = get_correctness_score(self.tokenizer, pred_t, exp_t)
        assert c >= 99.0

    def test_match_01(self):
        exp_t = "The sun is at the center of the solar system."
        pred_t = "the sun is a star at the center of the solar system"
        c = get_correctness_score(self.tokenizer, pred_t, exp_t)
        assert c >= 60

    def test_match_02(self):
        exp_t = "The sun is at the center of the solar system."
        pred_t = "Our star, the sun, is at the center of the solar system and the Earth is about 8 light minutes away!"
        c = get_correctness_score(self.tokenizer, pred_t, exp_t)
        assert c >= 70

    def test_mismatch_01(self):
        exp_t = "The sun is at the center of the solar system."
        pred_t = "It was the best of times, it was the worst of times!"
        c = get_correctness_score(self.tokenizer, pred_t, exp_t)
        assert c <= 30



import torch
from torch import nn


class ContrastClassifier(nn.Module):
    def __init__(self, scale: float):
        super(ContrastClassifier, self).__init__()
        self.scale = scale
        self.factor = nn.Parameter(torch.ones((1, 1)))
        self.prior_logit = nn.Parameter(torch.zeros((1, 1)))
        self.prior_weight_logit = nn.Parameter(torch.FloatTensor([[-5.0]]))

    def get_min_distance(self, rk: torch.Tensor, sk: torch.Tensor, keepdim=False):
        # rk: (B, num_r_keys, emb_size,)
        # sk: (B, num_s_keys, emb_size,)
        # Returns: (B,) or (B, 1,)
        diff = (sk[:, None, :, :] - rk[:, :, None, :]) * self.scale
        # diff_sq: (B, num_r_keys, num_s_keys, emb_size,)
        mean_diff_sq = torch.mean(diff.pow(2), dim=3)
        # diff_sq: (B, num_r_keys, num_s_keys,)
        s_mins = torch.amin(mean_diff_sq, dim=2)
        # s_mins: (B, num_r_keys,)
        return torch.amin(s_mins, dim=1, keepdim=keepdim)

    def forward(self, r_emb: torch.Tensor, s_emb: torch.Tensor):
        """
        r_emb: Tensor of size (batch_size, 1, emb_size,)
        s_emb: Tensor of size (batch_size, num_keys, emb_size,)
        """
        min_dist_sq = self.get_min_distance(r_emb, s_emb, keepdim=True)
        # min_dist_sq: (batch_size, 1,)
        nominal_p = torch.exp(-min_dist_sq * self.factor.pow(2))
        prior = torch.sigmoid(self.prior_logit)
        prior_weight = torch.sigmoid(self.prior_weight_logit)
        return prior * prior_weight + nominal_p * (1 - prior_weight)



import codecs
import os

from setuptools import setup, find_namespace_packages

packages = find_namespace_packages(include=['goodai.*'])
exec(open('goodai/ltm/version.py').read())
version = globals()['__version__']
print(f'Version: {version}')


def read_file(filename):
    with codecs.open(os.path.join(os.path.dirname(__file__), filename), 'r', 'utf-8') as file:
        return file.read()


setup(name='goodai-ltm',
      description='A text memory meant to be used with conversational language models.',
      long_description=read_file('README-Pypi.md'),
      long_description_content_type='text/markdown',
      url='https://github.com/GoodAI/goodai-ltm',
      version=version,
      packages=packages,
      package_data={'goodai.ltm.data': ['**/*.json']},
      install_requires=['torch>=1.8.0', 'pytest>=7.0.0', 'numpy>=1.19.0', 'transformers>=4.0.0',
                        'openai>=0.27.0', 'faiss-cpu', 'datasets', 'boto3', 'python-dotenv',
                        'sentence-transformers>=2.2.2']
)



from abc import abstractmethod
from typing import List, Tuple, Union, Callable

import numpy as np
import torch
from torch import nn
from tqdm import tqdm
from transformers import PreTrainedTokenizer

from goodai.helpers.tokenizer_helper import get_pad_token_id, get_model_inputs
from goodai.ltm.embeddings.base import BaseTextEmbeddingModel


class TrainableEmbeddingModel(BaseTextEmbeddingModel, nn.Module):
    """
    Abstract base class for locally trainable text embeddings.

    Trainable embedding models support multiple retrieval and storage embeddings for a query or passage.
    """

    def __init__(self, tokenizer: PreTrainedTokenizer):
        super(BaseTextEmbeddingModel, self).__init__()
        super(nn.Module, self).__init__()
        self.dummy = nn.Parameter()
        self.tokenizer = tokenizer
        self.pad_token_id = get_pad_token_id(tokenizer)

    def get_device(self):
        return self.dummy.device

    @abstractmethod
    def get_storage_emb(self, input_ids: torch.Tensor, attention_mask: torch.Tensor = None) -> torch.Tensor:
        """
        Returns an embedding of shape (batch_size, num_keys, emb_size,)
        """
        pass

    @abstractmethod
    def get_retrieval_emb(self, input_ids: torch.Tensor, token_lengths: torch.Tensor,
                          attention_mask: torch.Tensor = None) -> torch.Tensor:
        """
        Returns an embedding of shape (batch_size, emb_size,)
        """
        pass

    @abstractmethod
    def get_emb(self, input_ids: torch.Tensor, attention_mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Returns storage and retrieval embeddings.
        """
        pass

    @abstractmethod
    def get_lm_parameters(self):
        """
        :return: Fine-tuning parameters belonging to a pretrained language model.
        """
        pass

    @abstractmethod
    def get_added_parameters(self):
        """
        :return: New parameters that don't belong to the pretrained language model.
        """
        pass

    def get_storage_emb_for_ids(self, input_ids: List[List[int]]):
        device = self.get_device()
        pad_token_id = self.pad_token_id
        m_inputs = get_model_inputs(input_ids, pad_id=pad_token_id, device=device)
        return self.get_storage_emb(**m_inputs)

    def get_retrieval_emb_for_ids(self, input_ids: List[List[int]]):
        device = self.get_device()
        pad_token_id = self.pad_token_id
        m_inputs = get_model_inputs(input_ids, pad_id=pad_token_id, device=device, return_token_lengths=True,
                                    tokenizer=self.tokenizer)
        return self.get_retrieval_emb(**m_inputs)

    def _get_token_ids(self, texts: List[str]) -> List[List[int]]:
        tok = self.tokenizer
        return [tok.encode(text, add_special_tokens=False) for text in texts]

    def encode_in_batches(self, enc_fn: Callable, sentences: List[str],
                          add_special_tokens: bool = True,
                          batch_size: int = 50,
                          show_progress_bar: bool = False,
                          convert_to_tensor: bool = False,
                          return_token_lengths: bool = False,
                          detach_tensors: bool = True,
                          device: Union[str, torch.device] = None):
        device = device if device else self.get_device()
        t = self.tokenizer
        s_count = len(sentences)
        rng = range(0, s_count, batch_size)
        if show_progress_bar:
            rng = tqdm(rng, desc='Embeddings', unit='batch')
        keys_list = []
        for b0 in rng:
            b_sentences = sentences[b0:b0 + batch_size]
            input_ids_list = [t.encode(s, add_special_tokens=add_special_tokens) for s in b_sentences]
            model_inputs = get_model_inputs(input_ids_list, self.pad_token_id,
                                            return_token_lengths=return_token_lengths,
                                            tokenizer=self.tokenizer,
                                            device=device)
            keys = enc_fn(**model_inputs)
            if detach_tensors or not convert_to_tensor:
                keys = keys.detach()
            keys_list.append(keys)
        result = torch.cat(keys_list)
        if convert_to_tensor:
            return result
        else:
            return result.detach().cpu().numpy()

    def encode_queries(self, queries: List[str], batch_size: int = 64, show_progress_bar: bool = False,
                       convert_to_tensor: bool = False, detach_tensors: bool = True,
                       device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        return self.encode_in_batches(self.get_retrieval_emb, queries, batch_size=batch_size,
                                      show_progress_bar=show_progress_bar,
                                      convert_to_tensor=convert_to_tensor,
                                      detach_tensors=detach_tensors,
                                      return_token_lengths=True,
                                      device=device)

    def encode_corpus(self, passages: List[str], batch_size: int = 64, show_progress_bar: bool = False,
                      convert_to_tensor: bool = False, detach_tensors: bool = True,
                      device: Union[str, torch.device] = None) -> Union[np.ndarray, torch.Tensor]:
        return self.encode_in_batches(self.get_storage_emb, passages, batch_size=batch_size,
                                      show_progress_bar=show_progress_bar,
                                      convert_to_tensor=convert_to_tensor,
                                      detach_tensors=detach_tensors,
                                      return_token_lengths=False,
                                      device=device)



import pickle
from dataclasses import dataclass
from pathlib import Path

from goodai.ltm.mem.auto import AutoTextMemory
from goodai.ltm.mem.default import DefaultTextMemory
from goodai.ltm.persistence.abstract import MemoryPersistence


@dataclass
class WholeMemoryPicklePersistence(MemoryPersistence):
    """
    Saves and loads the whole memory as a pickle file. Platform dependent and heavily version dependent.
    Saves embedding model too so it's huge.
    """
    pickle_memory_filename: str = "memory.pickle"

    def save(self, memory: DefaultTextMemory, directory: Path):
        file = directory / self.pickle_memory_filename
        with open(file, "bw") as fd:
            pickle.dump(memory, fd, protocol=pickle.DEFAULT_PROTOCOL)

    def load(self, directory: Path, **kwargs) -> DefaultTextMemory:
        file = directory / self.pickle_memory_filename
        with open(file, "br") as fd:
            result = pickle.load(fd)
            if not isinstance(result, DefaultTextMemory):
                raise ValueError(f"Expected BaseTextMemory, got {type(result)}")
            return result

    def exists(self, directory: Path) -> bool:
        files = [
            directory / self.pickle_memory_filename,
        ]
        return all([file.exists() for file in files])


@dataclass
class TargetedMemoryPicklePersistence(MemoryPersistence):
    """
    Saves and loads the memory in a targeted way as a pickle file. Platform dependent and heavily version dependent,
    but only takes as much space as necessary.
    """
    chunk_queue_pickle_filename = "chunk_queue.pickle"
    vector_db_pickle_filename = "vector_db.pickle"

    def save(self, memory: DefaultTextMemory, directory: Path):
        chunk_queue = memory.chunk_queue
        file = directory / self.chunk_queue_pickle_filename
        with open(file, "bw") as fd:
            pickle.dump(chunk_queue, fd, protocol=pickle.DEFAULT_PROTOCOL)
        vector_db = memory.vector_db
        file = directory / self.vector_db_pickle_filename
        with open(file, "bw") as fd:
            pickle.dump(vector_db, fd, protocol=pickle.DEFAULT_PROTOCOL)

    def load(self, directory: Path, **kwargs) -> DefaultTextMemory:
        file = directory / self.chunk_queue_pickle_filename
        with open(file, "br") as fd:
            chunk_queue = pickle.load(fd)
        file = directory / self.vector_db_pickle_filename
        with open(file, "br") as fd:
            vector_db = pickle.load(fd)
        return AutoTextMemory.create(chunk_queue=chunk_queue, vector_db=vector_db, **kwargs)

    def exists(self, directory: Path) -> bool:
        files = [
            directory / self.chunk_queue_pickle_filename,
            directory / self.vector_db_pickle_filename,
        ]
        return all([file.exists() for file in files])



import codecs
import hashlib
import logging
import os
import pickle
import tempfile
import zipfile
import urllib.request
from typing import Any
from urllib.parse import urlparse
from tqdm import tqdm

_cache_dir = os.environ.get('GOODAI_CACHE', os.path.expanduser('~/.cache/goodai'))


def download_zip(url):
    hash_str = hashlib.sha256(url.encode('utf-8')).hexdigest()
    temp_dir = os.path.join(tempfile.gettempdir(), hash_str)

    if not os.path.exists(temp_dir):
        print(f'Downloading {url}')
        # Download the zip file
        zip_file, _ = urllib.request.urlretrieve(url)

        # Unzip the file into the temporary directory
        with zipfile.ZipFile(zip_file, 'r') as zip_ref:
            zip_ref.extractall(temp_dir)

        # Clean up the downloaded zip file
        os.remove(zip_file)
    else:
        print(f'File {url} already downloaded.')
    return temp_dir


def unpickle_downloaded_url(url: str) -> Any:
    for attempt in range(2):
        with open_url_as_file(url) as fd:
            try:
                return pickle.load(fd)
            except (EOFError, pickle.UnpicklingError):
                logging.warning(f'Discarding corrupted file downloaded from {url}')
        cache_file = get_cache_file_for_url(url)
        try:
            os.remove(cache_file)
        except (FileNotFoundError, PermissionError):
            # ignore
            pass
    raise SystemError('Failed to unpickle object')


def open_url_as_file(url, mode='rb'):
    file_name = url_as_file(url)
    return open(file_name, mode=mode)


def codecs_open_url_as_file(url, mode='r', encoding='utf-8'):
    file_name = url_as_file(url)
    return codecs.open(file_name, mode=mode, encoding=encoding)


def get_cache_file_for_url(url: str) -> str:
    if not os.path.exists(_cache_dir):
        os.makedirs(_cache_dir)

    # Convert the URL to a valid file name
    url_parts = urlparse(url)
    url_hash = hashlib.md5(url.encode('utf-8')).hexdigest()
    base_name = os.path.basename(url_parts.path)
    file_name = os.path.join(_cache_dir, url_hash + '_' + base_name)
    return file_name


def url_as_file(url) -> str:
    file_name = get_cache_file_for_url(url)
    progress_fn = file_name + '-in-progress'

    if not os.path.exists(file_name) or os.path.exists(progress_fn):
        logging.warning(f'Downloading {url}')
        with urllib.request.urlopen(url) as url_file:
            with open(progress_fn, 'w') as fd:
                fd.write('.')
            expected_size = url_file.getheader('Content-Length')
            if expected_size is not None:
                expected_size = int(expected_size)
            progress_bar = tqdm(total=expected_size, unit='B', unit_scale=True, desc='Download', leave=True)
            with open(file_name, 'wb') as local_file:
                while True:
                    chunk = url_file.read(100000)
                    if not chunk:
                        break
                    local_file.write(chunk)
                    progress_bar.update(len(chunk))
            os.remove(progress_fn)
    else:
        logging.info(f'Found cached version of {url}')
    return file_name




import math
from typing import List, Union, Any, Optional, Tuple, Callable
import numpy as np
import torch
from faiss import Index
from tqdm import tqdm
from transformers import PreTrainedTokenizer

from goodai.helpers.tokenizer_helper import get_pad_token_id, get_sentence_punctuation_ids
from goodai.ltm.embeddings.base import BaseTextEmbeddingModel
from goodai.ltm.mem.base import BaseReranker, BaseImportanceModel
from goodai.ltm.mem.chunk import Chunk, TextKeyType
from goodai.ltm.mem.rewrite_model import BaseRewriteModel
from goodai.ltm.reranking.base import BaseTextMatchingModel
from goodai.ltm.mem.chunk_queue import ChunkQueue, PassageInfo, ChunkExpansionOptions
from goodai.ltm.mem.config import TextMemoryConfig
from goodai.ltm.mem.mem_foundation import BaseTextMemoryFoundation, VectorDbType
from goodai.ltm.mem.simple_vector_db import SimpleVectorDb

_vector_db_type = Union[Index, SimpleVectorDb]


class DefaultTextMemory(BaseTextMemoryFoundation):
    def __init__(self, vector_db_type: VectorDbType, tokenizer: PreTrainedTokenizer,
                 emb_model: BaseTextEmbeddingModel, matching_model: Optional[BaseTextMatchingModel],
                 device: torch.device, config: TextMemoryConfig,
                 query_rewrite_model: Optional[BaseRewriteModel] = None,
                 memory_rewrite_model: Optional[BaseRewriteModel] = None,
                 reranker: Optional[BaseReranker] = None,
                 importance_model: Optional[BaseImportanceModel] = None,
                 chunk_queue: Optional[ChunkQueue] = None,
                 vector_db: Optional[_vector_db_type] = None,
                 ):
        cc = config.chunk_capacity
        cof = config.chunk_overlap_fraction
        if cof < 0 or cof > 0.5:
            raise ValueError(f'Invalid chunk overlap fraction: {cof}')
        ciao = cc - math.ceil(cc * cof)
        self.chunk_queue = chunk_queue or ChunkQueue(config.queue_capacity, cc, ciao)
        self.config = config
        self.device = device
        self.emb_model = emb_model
        self.matching_model = matching_model
        self.chunk_tokenizer = tokenizer
        self.bucket_capacity = config.chunk_capacity
        self.pad_token_id = get_pad_token_id(self.chunk_tokenizer)
        self.punctuation_ids = get_sentence_punctuation_ids(self.chunk_tokenizer, include_line_break=False)
        self.query_rewrite_model = query_rewrite_model
        self.memory_rewrite_model = memory_rewrite_model
        self.importance_model = importance_model
        self.reranker = reranker
        self.ce_options = ChunkExpansionOptions.from_config(tokenizer, config.chunk_expansion_config)
        has_matching_model = self.matching_model is not None
        super().__init__(vector_db_type, self.chunk_tokenizer, has_matching_model,
                         self.emb_model.get_num_storage_embeddings(),
                         self.emb_model.get_embedding_dim(),
                         config.chunk_capacity,
                         config.reranking_k_factor, config.max_query_length,
                         query_rewrite_model, reranker,
                         config.chunk_overlap_fraction, config.redundancy_overlap_threshold,
                         self.ce_options, device, vector_db=vector_db)

    def has_importance_model(self) -> bool:
        return self.importance_model is not None

    def get_tokenizer(self):
        return self.chunk_tokenizer

    def get_queue_capacity(self):
        return self.chunk_queue.get_capacity()

    def set_queue_capacity(self, num_buckets: int):
        self.chunk_queue.capacity = num_buckets

    def get_metadata(self, chunk_id: int) -> Optional[Any]:
        chunk = self.chunk_queue.get_chunk(chunk_id)
        if chunk is None:
            return None
        return chunk.metadata

    def get_chunk_text(self, chunk: Chunk) -> str:
        token_ids = self.chunk_queue.get_chunk_token_ids(chunk)
        return self.chunk_tokenizer.decode(token_ids, skip_special_tokens=True)

    def get_complete_passage(self, chunk: Chunk) -> PassageInfo:
        return self.chunk_queue.get_complete_passage(chunk, self.ce_options)

    def get_chunk(self, chunk_id: int) -> Chunk:
        return self.chunk_queue.get_chunk(chunk_id)

    def retrieve_chunk_sequences(self, chunks: List[Chunk]):
        return self.chunk_queue.retrieve_chunk_sequences_given_chunks(chunks)

    def get_text(self, text_key: TextKeyType) -> Optional[str]:
        token_ids = self.chunk_queue.get_sequence_token_ids(text_key)
        if token_ids is None:
            return None
        return self.chunk_tokenizer.decode(token_ids, skip_special_tokens=True)

    def get_retrieval_key_for_text(self, queries: List[str], show_progress_bar: bool = False) -> torch.Tensor:
        return self.emb_model.encode_queries(queries, convert_to_tensor=True, show_progress_bar=show_progress_bar)

    def predict_match(self, sentences: List[Tuple[str, str]], show_progress_bar: bool = False,
                      batch_size: int = 32) -> List[float]:
        return self.matching_model.predict(sentences, show_progress_bar=show_progress_bar,
                                           batch_size=batch_size)

    def _replace_or_add_text(self, cq_fn: Callable, text: str, metadata: Optional[Any] = None, rewrite: bool = False,
                             rewrite_context: Optional[str] = None, show_progress_bar: bool = False,
                             timestamp: Optional[float] = None, text_key: TextKeyType = None):
        if rewrite and not self.memory_rewrite_model:
            raise ValueError("For memory rewriting, a rewriting model must be provided")
        if rewrite and self.memory_rewrite_model:
            text = self.memory_rewrite_model.rewrite_memory(text, rewrite_context)
        importance = None
        if self.importance_model:
            importance = self.importance_model.get_importance(text)
        token_ids = self.chunk_tokenizer.encode(text, add_special_tokens=False)
        cq_params = dict(
            new_token_ids=token_ids,
            metadata=metadata,
            importance=importance,
            timestamp=timestamp,
        )
        if text_key is not None:
            cq_params['text_key'] = text_key
        removed_chunks, text_key = cq_fn(**cq_params)
        self._ensure_keys_added(show_progress_bar=show_progress_bar)
        removed_chunk_ids = [rb.chunk_id for rb in removed_chunks]
        if len(removed_chunk_ids) > 0:
            self.vector_db.remove_ids(np.array(removed_chunk_ids).astype(np.int64))
        return text_key

    def add_text(self, text: str, metadata: Optional[dict] = None, rewrite: bool = False,
                 rewrite_context: Optional[str] = None, show_progress_bar: bool = False,
                 timestamp: Optional[float] = None) -> TextKeyType:
        return self._replace_or_add_text(self.chunk_queue.add_sequence,
                                         text=text, metadata=metadata, rewrite=rewrite,
                                         rewrite_context=rewrite_context, show_progress_bar=show_progress_bar,
                                         timestamp=timestamp)

    def replace_text(self, text_key: TextKeyType, text: str, metadata: Optional[dict] = None,
                     rewrite: bool = False, rewrite_context: Optional[str] = None,
                     show_progress_bar: bool = False, timestamp: Optional[float] = None) -> TextKeyType:
        return self._replace_or_add_text(self.chunk_queue.replace_sequence,
                                         text=text, metadata=metadata, rewrite=rewrite,
                                         rewrite_context=rewrite_context, show_progress_bar=show_progress_bar,
                                         timestamp=timestamp, text_key=text_key)

    def delete_text(self, text_key: TextKeyType, show_progress_bar: bool = False) -> TextKeyType:
        return self.replace_text(text_key, "", show_progress_bar=show_progress_bar)

    def add_separator(self):
        self.chunk_queue.add_separator()

    def is_empty(self) -> bool:
        return len(self.chunk_queue.token_ids) == 0

    def retrieve_all_text(self) -> str:
        token_ids = self.chunk_queue.get_latest_token_ids(max_num_tokens=None)
        return self.chunk_tokenizer.decode(token_ids, skip_special_tokens=True)

    def retrieve_all_chunks(self) -> List[str]:
        chunk_list = self.chunk_queue.get_chunk_sequences()
        tok = self.chunk_tokenizer
        return [tok.decode(seq, skip_special_tokens=True) for seq in chunk_list]

    def get_all_chunks(self) -> List[Chunk]:
        return self.chunk_queue.get_all_chunks()

    def _ensure_keys_added(self, batch_size=50, show_progress_bar: bool = False):
        picked_chunks, token_id_matrix = self.chunk_queue.get_chunks_for_indexing()
        emb_model = self.emb_model
        sk_list = []
        num_sequences = len(token_id_matrix)
        rng = range(0, num_sequences, batch_size)
        if show_progress_bar:
            rng = tqdm(rng, desc='Storage', unit='batch')
        for i in rng:
            token_id_batch = token_id_matrix[i:i + batch_size]
            text_batch = self.chunk_tokenizer.batch_decode(token_id_batch, skip_special_tokens=True)
            sk_batch = emb_model.encode_corpus(text_batch, convert_to_tensor=True,
                                               show_progress_bar=False, batch_size=batch_size)
            if sk_batch.size(0) != len(text_batch):
                raise RuntimeError(f'Number of storage embeddings returned by embedding model is {sk_batch.size(0)}, '
                                   f'while the number of encoded texts is {len(text_batch)}')
            sk_list.append(sk_batch.detach())
        if len(sk_list) > 0:
            sk_all = torch.cat(sk_list)
            num_chunks, num_sk = sk_all.size(0), sk_all.size(1),
            sk_all = sk_all.view(num_chunks * num_sk, -1)
            sk_all_np = sk_all.cpu().numpy().astype(np.float32)
            b_indexes = []
            for chunk in picked_chunks:
                chunk.update_indexed_state()
                b_indexes.extend([chunk.chunk_id] * num_sk)
            b_indexes_np = np.array(b_indexes).astype(np.int64)
            self.vector_db.add_with_ids(sk_all_np, b_indexes_np)

    def clear(self):
        self.chunk_queue.flush()
        self.vector_db.reset()



import tempfile
import unittest
from datetime import datetime
from pathlib import Path

from goodai.ltm.embeddings.auto import AutoTextEmbeddingModel
from goodai.ltm.mem.auto import AutoTextMemory
from goodai.ltm.mem.base import BaseTextMemory
from goodai.ltm.mem.default import DefaultTextMemory
from goodai.ltm.persistence.abstract import MemoryPersistence
from goodai.ltm.persistence.json_persistence import JsonNpzMemoryPersistence, JsonMemoryPersistence
from goodai.ltm.persistence.msgpack_persistence import MsgNpzPackMemoryPersistence, MsgPackMemoryPersistence
from goodai.ltm.persistence.pickle_persistence import WholeMemoryPicklePersistence, TargetedMemoryPicklePersistence


class TestPersistence(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._lr_emb_model = AutoTextEmbeddingModel.shared_pretrained('em-MiniLM-p1-01')
        cls._text = "Earth has a dynamic atmosphere, which sustains Earth's surface conditions and protects " \
                    "it from most meteoroids and UV-light at entry. It has a composition of primarily nitrogen " \
                    "and oxygen. Water vapor is widely present in the atmosphere, forming clouds that cover most " \
                    "of the planet. The water vapor acts as a greenhouse gas and, together with other greenhouse " \
                    "gases in the atmosphere, particularly carbon dioxide (CO2), creates the conditions for both " \
                    "liquid surface water and water vapor to persist via the capturing of energy from the Sun's " \
                    "light. This process maintains the current average surface temperature of 14.76 °C, at " \
                    "which water is liquid under atmospheric pressure. Differences in the amount of captured energy " \
                    "between geographic regions (as with the equatorial region receiving more sunlight than the " \
                    "polar regions) drive atmospheric and ocean currents, producing a global climate system with " \
                    "different climate regions, and a range of weather phenomena such as precipitation, allowing " \
                    "components such as nitrogen to cycle."

        cls._kwargs = {
            "emb_model": cls._lr_emb_model
            # "emb_model": 'em-MiniLM-p1-01'
        }

    def test_saving_pickle(self):
        mem = self.do_saving(WholeMemoryPicklePersistence(), **self._kwargs)

    def test_saving_pickle_targeted(self):
        mem = self.do_saving(TargetedMemoryPicklePersistence(), **self._kwargs)

    def test_saving_json_npz(self):
        mem = self.do_saving(JsonNpzMemoryPersistence(), **self._kwargs)

    def test_saving_json(self):
        mem = self.do_saving(JsonMemoryPersistence(), **self._kwargs)

    def test_saving_msgpack_npz(self):
        mem = self.do_saving(MsgNpzPackMemoryPersistence(), **self._kwargs)

    def test_saving_msgpack(self):
        mem = self.do_saving(MsgPackMemoryPersistence(), **self._kwargs)

    def do_saving(self, memory_persistence: MemoryPersistence, **kwargs):
        mem: DefaultTextMemory = AutoTextMemory.create(**kwargs)
        mem.add_text(self._text, metadata={'foo': 'bar', "timestamp": datetime.now()})
        directory = tempfile.TemporaryDirectory().name
        path = Path(directory)
        path.mkdir(parents=True, exist_ok=True)

        self.verify_memory(mem)

        memory_persistence.save(mem, path)

        loaded_memory = memory_persistence.load(path, **kwargs)

        assert loaded_memory.chunk_queue == mem.chunk_queue
        assert loaded_memory.vector_db == mem.vector_db

        self.verify_memory(loaded_memory)

        return loaded_memory

    def verify_memory(self, mem: BaseTextMemory):
        query = "Is water vapor widely present in the atmosphere?"
        r_memories = mem.retrieve(query, k=2)
        r_combined_text = '\n'.join([rm.passage for rm in r_memories])
        assert "Water vapor is widely present in the atmosphere" in r_combined_text



import threading
import openai
from typing import Optional
from goodai.text_gen.base import BaseTextGenerationModel

_openai_lock = threading.Lock()


class OpenAICompletionModel(BaseTextGenerationModel):
    def __init__(self, model_name: str, api_key: str = None, temperature: Optional[float] = None,
                 max_tokens: Optional[int] = None):
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.api_key = api_key
        self.model_name = model_name

    def generate(self, prompt: str) -> str:
        with _openai_lock:
            if self.api_key:
                openai.api_key = self.api_key
            response = openai.Completion.create(model=self.model_name, prompt=prompt, temperature=self.temperature,
                                                max_tokens=self.max_tokens)
            return response['choices'][0]['text'].strip()


class OpenAIChatCompletionModel(BaseTextGenerationModel):
    def __init__(self, model_name: str, api_key: str = None, temperature: Optional[float] = None,
                 max_tokens: Optional[int] = None):
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.api_key = api_key
        self.model_name = model_name

    def generate(self, prompt: str) -> str:
        with _openai_lock:
            if self.api_key:
                openai.api_key = self.api_key
            messages = [
                {"role": "user", "content": prompt}
            ]
            response = openai.ChatCompletion.create(model=self.model_name, messages=messages,
                                                    temperature=self.temperature,
                                                    max_tokens=self.max_tokens)
            return response['choices'][0]['message']['content'].strip()



import abc
import enum
import io
import json
import math
import sys
from dataclasses import dataclass
from typing import List, Union, Set, Optional, Tuple

import faiss
import numpy as np
import torch
from transformers import PreTrainedTokenizer

from goodai.ltm.mem.base import BaseTextMemory, RetrievedMemory, BaseReranker
from goodai.ltm.mem.chunk import Chunk
from goodai.ltm.mem.chunk_queue import PassageInfo, ChunkExpansionOptions
from goodai.ltm.mem.rewrite_model import BaseRewriteModel
from goodai.ltm.mem.simple_vector_db import SimpleVectorDb

_vector_db_type = Union[faiss.Index, SimpleVectorDb]


class VectorDbType(enum.Enum):
    SIMPLE = 0,
    FAISS_FLAT_L2 = 1,


@dataclass
class RetrievedChunk:
    chunk: Chunk
    distance: float
    passage: PassageInfo
    confidence: Optional[float] = None

    def with_confidence(self, confidence: float) -> 'RetrievedChunk':
        return RetrievedChunk(chunk=self.chunk, distance=self.distance,
                              passage=self.passage,
                              confidence=confidence)

    def sort_key(self):
        return self.distance if self.confidence is None else -self.confidence

    @staticmethod
    def has_overlap(included_indexes: Set[int], p_from: int, p_to: int,
                    overlap_threshold: float):
        if p_from not in included_indexes and (p_to - 1) not in included_indexes:
            return False
        if p_to <= p_from:
            return False
        intersection = included_indexes.intersection(range(p_from, p_to))
        len_inter = len(intersection)
        if len_inter <= 0:
            return False
        overlap_fraction = len_inter / (p_to - p_from)
        return overlap_fraction >= overlap_threshold

    @classmethod
    def remove_duplicates_and_overlaps(cls, items: List['RetrievedChunk'],
                                       overlap_threshold: float, max_count: int) -> List['RetrievedChunk']:
        id_set = set()
        result = []
        included_indexes: Set[int] = set()
        for item in items:
            chunk = item.chunk
            if chunk is None:
                continue
            chunk_id = chunk.chunk_id
            if chunk_id in id_set:
                continue
            id_set.add(chunk_id)
            passage = item.passage
            p_from = passage.fromIndex
            p_to = passage.toIndex
            if cls.has_overlap(included_indexes, p_from, p_to, overlap_threshold):
                continue
            result.append(item)
            if len(result) >= max_count:
                break
            included_indexes.update(range(p_from, p_to))
        return result

    @classmethod
    def reduce(cls, items: List['RetrievedChunk'], overlap_threshold: float,
               expected_non_redundant_count: int) -> List['RetrievedChunk']:
        # Does not remove overlapping passages.
        # Returns as many passages as would be needed to get expected_non_redundant_count
        # after removal of overlapping passages.
        id_set = set()
        result = []
        included_indexes: Set[int] = set()
        count = 0
        for item in items:
            chunk = item.chunk
            if chunk is None:
                continue
            chunk_id = chunk.chunk_id
            if chunk_id in id_set:
                continue
            id_set.add(chunk_id)
            passage = item.passage
            p_from = passage.fromIndex
            p_to = passage.toIndex
            result.append(item)
            if not cls.has_overlap(included_indexes, p_from, p_to, overlap_threshold):
                count += 1
                if count >= expected_non_redundant_count:
                    break
            included_indexes.update(range(p_from, p_to))
        return result


class BaseTextMemoryFoundation(BaseTextMemory):
    def __init__(self,
                 vector_db_type: VectorDbType, tokenizer: PreTrainedTokenizer, has_match_prob_model: bool,
                 num_storage_embeddings: int, emb_dim: int,
                 chunk_capacity: int, reranking_k_factor: float,
                 max_query_length: Optional[int],
                 query_rewrite_model: BaseRewriteModel, reranker: BaseReranker,
                 overlap_fraction: float,
                 overlap_threshold: float, chunk_expansion_options: ChunkExpansionOptions,
                 device: torch.device,
                 max_expansion_top_k_factor: int = 200,
                 vector_db: Optional[_vector_db_type] = None
                 ):
        super().__init__()
        if overlap_fraction < 0 or overlap_fraction > 0.5:
            raise ValueError(f'Invalid chunk overlap fraction: {overlap_fraction}')
        if overlap_threshold <= 0 or overlap_threshold > 1.0:
            raise ValueError(f'Invalid redundancy overlap threshold: {overlap_threshold}')
        if reranking_k_factor < 1:
            raise ValueError('reranking_k_factor cannot be less than 1')
        max_st = chunk_expansion_options.maxSideTokens
        if max_st < chunk_expansion_options.minSideTokens:
            raise ValueError('Invalid chunk expansion configuration: The maximum number of extra side tokens '
                             'cannot be less than the minimum number.')
        max_added_chunks = max_st * 2 / chunk_capacity
        max_chunks_overlap = (max_added_chunks + 2) / (1 - overlap_fraction)
        self.expansion_top_k_factor = overlap_threshold + (1 - overlap_threshold) * (1 + max_chunks_overlap)
        if self.expansion_top_k_factor > max_expansion_top_k_factor:
            raise ValueError('Excessive chunk expansion configuration. Increase the chunk capacity or the redundancy '
                             'overlap threshold, or decrease the maximum extra side tokens or the chunk '
                             'overlap fraction.')
        self.chunk_expansion_options = chunk_expansion_options
        self.chunk_capacity = chunk_capacity
        self.overlap_threshold = overlap_threshold
        self.overlap_fraction = overlap_fraction
        self.max_query_length = max_query_length
        self.reranking_k_factor = reranking_k_factor
        self.num_storage_embeddings = num_storage_embeddings
        self.has_match_prob_model = has_match_prob_model
        self.emb_dim = emb_dim
        self.vector_db = vector_db or self.create_vector_db(vector_db_type, emb_dim)
        self.device = device
        self.chunk_tokenizer = tokenizer
        self.query_rewrite_model = query_rewrite_model
        self.reranker = reranker

    @staticmethod
    def create_vector_db(vector_db_type: VectorDbType, emb_dim: int) -> _vector_db_type:
        if vector_db_type == VectorDbType.SIMPLE:
            return SimpleVectorDb()
        elif vector_db_type == VectorDbType.FAISS_FLAT_L2:
            return faiss.IndexIDMap(faiss.IndexFlatL2(emb_dim))
        else:
            raise ValueError(f'Unrecognized vector DB type: {vector_db_type}')

    def get_tokenizer(self):
        return self.chunk_tokenizer

    @abc.abstractmethod
    def retrieve_chunk_sequences(self, chunks: List[Chunk]):
        pass

    @abc.abstractmethod
    def get_retrieval_key_for_text(self, queries: List[str], show_progress_bar: bool = False) -> torch.Tensor:
        pass

    def predict_match(self, sentences: List[Tuple[str, str]], show_progress_bar: bool = False,
                      batch_size: int = 32) -> List[float]:
        pass

    @abc.abstractmethod
    def get_metadata(self, chunk_id: int) -> Optional[dict]:
        pass

    @abc.abstractmethod
    def get_all_chunks(self) -> List[Chunk]:
        pass

    @abc.abstractmethod
    def get_chunk_text(self, chunk: Chunk) -> str:
        pass

    @abc.abstractmethod
    def get_chunk(self, chunk_id: int) -> Chunk:
        pass

    @abc.abstractmethod
    def get_complete_passage(self, chunk: Chunk) -> PassageInfo:
        pass

    def dump(self, stream: io.TextIOBase = sys.stdout):
        chunks = self.get_all_chunks()
        stream.write('| Id | Metadata | Importance | Content |\n')
        stream.write('| ----- | -------- | ---------- | ------- |\n')
        for chunk in chunks:
            chunk_text = self.get_chunk_text(chunk)
            ct_js = json.dumps(chunk_text)
            importance_text = 'None' if chunk.importance is None else f'{chunk.importance:.2g}'
            stream.write(f'| {chunk.chunk_id} | {chunk.metadata} | | {importance_text} | {ct_js} |')
            stream.write('\n')
        stream.flush()

    @staticmethod
    def _distance_to_relevance(distance: float, confidence: Optional[float]) -> float:
        if confidence is not None:
            return confidence
        # Assuming embeddings are always unit vectors.
        # Also assuming vector DB (like FAISS) returns squared Euclidean distance.
        # So distance can go from 0 to 4.
        return 1 - distance / 4.0

    def _retrieve_for_processed_r_chunks(self, processed_r_chunks: List[RetrievedChunk], k: int) -> \
            List[RetrievedMemory]:
        if not self.reranker:
            processed_r_chunks = processed_r_chunks[:k]
        has_pm = self.has_match_prob_model
        passage_info_list = [r_chunk.passage for r_chunk in processed_r_chunks]
        sequences = [r_chunk.passage.tokenIds for r_chunk in processed_r_chunks]
        retrieved_texts = self.chunk_tokenizer.batch_decode(sequences, skip_special_tokens=True)
        result = []
        for p_info, r_chunk, r_text in zip(passage_info_list, processed_r_chunks, retrieved_texts):
            confidence = r_chunk.confidence
            chunk = r_chunk.chunk
            metadata = chunk.metadata
            importance = chunk.importance
            distance = r_chunk.distance
            relevance = self._distance_to_relevance(distance, confidence)
            if not has_pm:
                confidence = None
            result.append(RetrievedMemory(passage=r_text.strip(), passage_info=p_info,
                                          timestamp=chunk.timestamp,
                                          distance=distance, relevance=relevance,
                                          textKeys=list(chunk.associated_keys),
                                          confidence=confidence,
                                          metadata=metadata, importance=importance))
        if self.reranker:
            result = self.reranker.rerank(result, self)
            result = result[:k]
        return result

    def _multi_retrieve_for_r_chunks(self, queries: List[str], prelim_r_chunks: List[List[RetrievedChunk]],
                                     reranking_top_k: int, k: int,
                                     show_progress_bar: bool = False) -> List[List[RetrievedMemory]]:
        # At this point:
        # - No chunks are None
        # - Duplicate chunk IDs have already been removed in prelim_r_chunks
        # - Overlapping chunks are yet to be removed
        # - prelim_r_chunks is ordered by distance
        if self.has_match_prob_model:
            m_sentences = []
            m_indexes = []
            reduced_r_chunks = []
            for i, (query, row_r_chunks) in enumerate(zip(queries, prelim_r_chunks)):
                row_reduced_r_chunks = RetrievedChunk.reduce(row_r_chunks, self.overlap_threshold,
                                                             reranking_top_k)
                reduced_r_chunks.append(row_reduced_r_chunks)
                prelim_chunks = [r_chunk.chunk for r_chunk in row_reduced_r_chunks]
                chunk_sequences: List[List[int]] = self.retrieve_chunk_sequences(prelim_chunks)
                chunk_texts = self.chunk_tokenizer.batch_decode(chunk_sequences, skip_special_tokens=True)
                for ct in chunk_texts:
                    m_sentences.append((query, ct,))
                    m_indexes.append(i)
            match_probs = self.predict_match(m_sentences, show_progress_bar=show_progress_bar)
            chunk_probs_list = [[] for _ in range(len(queries))]
            for m_index, match_prob in zip(m_indexes, match_probs):
                chunk_probs_list[m_index].append(match_prob)
            processed_r_chunks = []
            for chunk_probs, row_r_chunks in zip(chunk_probs_list, reduced_r_chunks):
                new_row_r_chunks = [r_chunk.with_confidence(cp) for cp, r_chunk in zip(chunk_probs, row_r_chunks)]
                new_row_r_chunks.sort(key=RetrievedChunk.sort_key)
                processed_r_chunks.append(new_row_r_chunks)
        else:
            processed_r_chunks = prelim_r_chunks
        result = []
        for row_r_chunks in processed_r_chunks:
            # Removal of overlapping passages
            row_r_chunks = RetrievedChunk.remove_duplicates_and_overlaps(row_r_chunks, self.overlap_threshold,
                                                                         reranking_top_k)
            # At this point there are at most reranking_top_k chunks per row
            result.append(self._retrieve_for_processed_r_chunks(row_r_chunks, k=k))
        return result

    def _multi_retrieve(self, queries: List[str], distances: np.ndarray, indexes: np.ndarray,
                        reranking_top_k: int, expansion_top_k: int, k: int,
                        show_progress_bar: bool = False) -> List[List[RetrievedMemory]]:
        # len(queries) == batch_size
        # distances, indexes: (batch_size, downstream_top_k)
        prelim_r_chunks = []
        batch_size = len(queries)
        assert batch_size == distances.shape[0] == indexes.shape[0], \
            f'batch_size={batch_size}, distances.shape={distances.shape}, indexes.shape={indexes.shape}'
        for i in range(batch_size):
            row_d = distances[i]
            row_i = indexes[i]
            # Remove duplicate and "not found" chunk IDs
            distinct_d_i = []
            id_set = set()
            for distance, chunk_id in zip(row_d, row_i):
                if chunk_id in id_set or chunk_id < 0:
                    continue
                id_set.add(chunk_id)
                distinct_d_i.append((distance, chunk_id,))
            # Without duplicates, we only need at most expansion_top_k chunks at this point
            distinct_d_i = distinct_d_i[:expansion_top_k]

            # Retrieve chunk objects from queue
            row_chunks = [self.get_chunk(chunk_id) for _, chunk_id in distinct_d_i]
            row_passages = [self.get_complete_passage(chunk) for chunk in row_chunks]
            row_r_chunks = [RetrievedChunk(chunk, distance, passage)
                            for chunk, passage, (distance, _) in
                            zip(row_chunks, row_passages, distinct_d_i)]
            prelim_r_chunks.append(row_r_chunks)
        return self._multi_retrieve_for_r_chunks(queries, prelim_r_chunks, reranking_top_k, k,
                                                 show_progress_bar=show_progress_bar)

    def retrieve_multiple(self, queries: List[str], k: int, rewrite: bool = False, show_progress_bar: bool = False,
                          **kwargs) -> List[List[RetrievedMemory]]:
        if k <= 0:
            raise ValueError('k must be greater than zero')
        if rewrite and not self.query_rewrite_model:
            raise ValueError("For query rewriting, a rewriting model must be provided")
        if rewrite and self.query_rewrite_model:
            queries = [self.query_rewrite_model.rewrite_query(q) for q in queries]
        if self.max_query_length is not None:
            queries = self._truncate_queries(queries, self.max_query_length)
        rk = self.get_retrieval_key_for_text(queries, show_progress_bar=show_progress_bar)
        batch_size, num_rk, emb_size = rk.size(0), rk.size(1), rk.size(2),
        if num_rk != 1:
            raise RuntimeError('Memory does not support multiple retrieval embeddings')
        n_queries = len(queries)
        if batch_size != n_queries:
            raise RuntimeError(f'Batch returned by embeddings model is of shape {rk.shape} '
                               f'while the number of queries is {n_queries}')
        rk_np = rk.view(batch_size, emb_size).detach().cpu().numpy()
        reranking_top_k = k
        if self.has_match_prob_model or self.reranker is not None:
            reranking_top_k = round(reranking_top_k * self.reranking_k_factor)

        # Extra items retrieved, anticipating possible overlaps
        expansion_top_k = 1 + math.ceil((reranking_top_k - 1) * self.expansion_top_k_factor)

        # The vector_db can return multiple entries with the same ID if chunks are associated
        # with multiple embeddings
        coverage_top_k = 1 + (expansion_top_k - 1) * self.num_storage_embeddings

        # Call to vector database
        distances, indexes = self.vector_db.search(rk_np, k=coverage_top_k)

        # Assumes vector_db returns ordered results
        assert distances.shape[0] == indexes.shape[0] == batch_size
        return self._multi_retrieve(queries, distances, indexes, reranking_top_k, expansion_top_k, k,
                                    show_progress_bar=show_progress_bar)

    def _truncate_queries(self, queries: List[str], max_query_length: int):
        if max_query_length <= 0:
            raise ValueError(f'max_query_length={max_query_length} is invalid')
        ct = self.chunk_tokenizer
        tokenization = ct.batch_encode_plus(queries, add_special_tokens=False,
                                            return_attention_mask=False)
        token_ids: List[List[int]] = tokenization['input_ids']
        result = []
        for seq, query in zip(token_ids, queries):
            if len(seq) > max_query_length:
                truncated_seq = seq[-max_query_length:]
                result.append(ct.decode(truncated_seq, skip_special_tokens=True))
            else:
                result.append(query)
        return result
